#! /usr/bin/env python
#
# Functions for debugging, such as console tracing. This is for intended
# for verbose tracing not suitable for the logging facility.
#
# Notes:
# - These are no-op's unless __debug__ is True.
# - Running python with the -O (optimized) option ensures that __debug__ is False.
# - So that other local packages can use tracing freely, this only
#   imports standard packages. In particular, system.py is not imported,
#   so functionality must be reproduced here (e.g., _to_utf8).
# - Gotta hate Pythonista's who prevailed in Python3 breaking lots of Python2
#   code just for the sake of simplicity, a la manera moronista (i.e., only one moronic way to do things)!
# - A few functions from system are re-implemented here without tracing to
#   avoid circular dependencies (e.g., _to_utf8).
# - For f-string syntax, see following:
#   https://docs.python.org/3/tutorial/inputoutput.html
#   https://www.python.org/dev/peps/pep-0498
#
# Usage examples:
#
#   - casual usage (quick-n-dirty debugging)
#
#     from mezcla import debug
#     debug.trace_expr(4, dir())
#
#   - long-term usuage (e.g., symbolic constants for consistency across scripts)
#
#     from mezcla import debug
#     TL = debug.TL
#     debug.trace(TL.DEFAULT, "Shown by default")
#     ...
#     debug.trace(TL.USUAL, f"Look ma: TL.VERBOSE={int(TL.VERBOSE)}")
#
#   - use DEBUG_LEVEL env. var. to set level
#
#     python -c 'from mezcla import debug; debug.trace(debug.DEFAULT + 1, "Not visible")'
#     DEBUG_LEVEL=3 python -c 'from mezcla import debug; debug.trace(3, "Visible")'
#
# TODO1:
# - Add sanity check to trace_fmt for when keyword in kaargs unused.
#
# TODO:
# - * Add sanity checks for unused environment variables specified on command line (e.g., FUBAR=1 python script.py ...)!
# - Rename as debug_utils so clear that non-standard package.
# - Add exception handling throughout (e.g., more in trace_object).
# - Apply format_value consistently.
#
#

"""Debugging functions (e.g., tracing)"""

# Standard packages
import atexit
from datetime import datetime
import enum
import inspect
from itertools import zip_longest
import logging
import os
from pprint import pprint
import re
## OLD: from xml.dom.minidom import Element
import six
import sys
import time

# Local packages
# note: The following redefines sys.version_info to be python3 compatible;
# this is used in _to_utf8, which should be reworked via six-based wrappers.
import mezcla.sys_version_info_hack      # pylint: disable=unused-import


# Constants for pre-defined tracing levels
#
class TraceLevel(enum.IntEnum):
    """Constants for use in tracing"""
    ALWAYS = 0
    ERROR = 1
    WARNING = 2                         # typically always shown
    DEFAULT = WARNING
    USUAL = 3                           # usual in sense of debugging purposes
    DETAILED = 4
    VERBOSE = 5
    QUITE_DETAILED = 6                  # detailed I/O
    QUITE_VERBOSE = 7
    MOST_DETAILED = 8
    MOST_VERBOSE = 9                    # for internal debugging
#
TL = TraceLevel
ALWAYS = TL.ALWAYS
ERROR = TL.ERROR
WARNING = TL.WARNING
DEFAULT = TL.DEFAULT
USUAL = TL.USUAL
DETAILED = TL.DETAILED
VERBOSE = TL.VERBOSE
QUITE_DETAILED = TL.QUITE_DETAILED
QUITE_VERBOSE = TL.QUITE_VERBOSE
MOST_DETAILED = TL.MOST_DETAILED
MOST_VERBOSE = TL.MOST_VERBOSE

# Other constants
UTF8 = "UTF-8"
STRING_TYPES = six.string_types
# note: INDENT0 is left margin and INDENT1 is normal indent
INDENT0 = ""
INDENT1 = "    "
INDENT = INDENT1
MISSING_LINE = "???"

# Globals
max_trace_value_len = 512

if __debug__:    

    # Initialize debug tracing level
    # TODO: mark as "private" (e.g., trace_level => _trace_level)
    DEBUG_LEVEL_LABEL = "DEBUG_LEVEL"
    trace_level = TL.DEFAULT            # typically same as TL.WARNING (2); TODO: global_trace_level
    output_timestamps = False           # prefix output with timestamp
    last_trace_time = time.time()       # timestamp from last trace
    use_logging = False                 # traces via logging (and stderr)
    debug_file = None                   # file for log output
    debug_file_hack = False             # work around concurrent writes by reopening after each trace
    para_mode_tracing = False           # multiline tracing adds blank line (e.g., for para-mode grep)
    #
    try:
        trace_level_text = os.environ.get(DEBUG_LEVEL_LABEL, "")
        if trace_level_text.strip():
            trace_level = int(trace_level_text)
    except:
        ## sys.stderr.write("Warning: Unable to set tracing level from {v}: {exc}\n".
        ##                  format(v=DEBUG_LEVEL_LABEL, exc=sys.exc_info()))
        pass


    def set_level(level):
        """Set new trace level"""
        global trace_level
        trace_level = level
        return


    def get_level():
        """Get current tracing level"""
        # Note: ensures result is integer (not enum)
        # EX: (get_level() >= 0)
        # EX: type(get_level() == int)
        ## OLD:
        ## ## global trace_level
        ## return trace_level
        level = 0
        try:
            ## OLD: assertion(isinstance(level, int))
            level = int(trace_level)
        except:
            _print_exception_info("get_level")
        return level

    def get_output_timestamps():
        """Return whether outputting timestamps"""
        return output_timestamps


    def set_output_timestamps(do_output_timestamps):
        """Enable for disable the outputting of timestamps"""
        global output_timestamps
        output_timestamps = do_output_timestamps


    def _to_utf8(text):
        """Convert TEXT to UTF-8 (e.g., for I/O)"""
        # Note: version like one from system.py to avoid circular dependency
        result = text
        if ((sys.version_info.major < 3) and (isinstance(text, unicode))):  # pylint: disable=undefined-variable
            result = result.encode("UTF-8", 'ignore')
        return result


    def _to_unicode(text, encoding=None):
        """Ensure TEXT in ENCODING is Unicode, such as from the default UTF8"""
        # TODO: rework from_utf8 in terms of this
        if not encoding:
            encoding = UTF8
        result = text
        if ((sys.version_info.major < 3) and (not isinstance(result, unicode))): # pylint: disable=undefined-variable
            result = result.decode(encoding, 'ignore')
        return result


    def _to_string(text):
        """Ensure TEXT is a string type"""
        result = text
        if (not isinstance(result, STRING_TYPES)):
            # Values are coerced using % operator for proper Unicode handling,
            # except for tuples which are converted recursively. This avoids a
            # type error due to arguments not being converted (e.g., second 
            # tuple constituent, etc.), as in ("%s" % (9, 1)).
            if isinstance(result, tuple):
                result = "(" + ", ".join([_to_string(v) for v in result]) + ")"
            else:
                result = "%s" % result
        return result

    def do_print(text, end=None):
        """Print TEXT to stderr and optionally to DEBUG_FILE"""
        print(text, file=sys.stderr, end=end)
        if debug_file:
            print(text, file=debug_file, end=end)
    
    def trace(level, text, empty_arg=None, no_eol=False, indentation=None):
        """Print TEXT if at trace LEVEL or higher, including newline unless SKIP_NEWLINE"""
        # TODO: add option to use format_value
        # Note: trace should not be used with text that gets formatted to avoid
        # subtle errors
        ## DEBUG: sys.stderr.write("trace({l}, {t})\n".format(l=level, t=text))
        if (trace_level >= level):
            if indentation is None:
                indentation = INDENT0
            # Prefix trace with timestamp w/o date
            if output_timestamps:
                # Get time-proper from timestamp (TODO: find standard way to do this)
                timestamp_time = re.sub(r"^\d+-\d+-\d+\s*", "", timestamp())
                if detailed_debugging():
                    global last_trace_time
                    diff = round(1000.0 * (time.time() - last_trace_time), 3)
                    timestamp_time += f" diff={diff}ms"
                    last_trace_time = time.time()
                do_print(indentation + "[" + timestamp_time + "]", end=": ")
            # Print trace, converted to UTF8 if necessary (Python2 only)
            # TODO: add version of assertion that doesn't use trace or trace_fmtd
            ## TODO: assertion(not(re.search(r"{\S*}", text)))
            end = "\n" if (not no_eol) else ""
            do_print(indentation + _to_utf8(text), end=end)
            if use_logging:
                # TODO: see if way to specify logging terminator
                logging.debug(indentation + _to_utf8(text))
            if debug_file_hack:
                reopen_debug_file()
        if empty_arg is not None:
            sys.stderr.write("Error: trace only accepts two positional arguments (was trace_expr intended?)\n")
        return


    def trace_fmtd(level, text, **kwargs):
        """Print TEXT with formatting using optional format KWARGS if at trace LEVEL or higher, including newline"""
        # Note: To avoid interpolated text as being interpreted as variable
        # references, this function does the formatting.
        # TODO: weed out calls that use (level, text.format(...)) rather than (level, text, ...)
        if (trace_level >= level):
            max_len = kwargs.get('_max_len') or kwargs.get('max_len')
            try:
                try:
                    # TODO: add version of assertion that doesn't use trace or trace_fmtd
                    ## TODO: assertion(re.search(r"{\S*}", text))
                    ## OLD: assertion("{" in text)
                    ## OLD: trace(level, text.format(**kwargs))
                    ## OLD: kwargs_unicode = {k: _to_unicode(_to_string(v)) for (k, v) in list(kwargs.items())}
                    kwargs_unicode = {k: format_value(_to_unicode(_to_string(v)), max_len=max_len)
                                      for (k, v) in list(kwargs.items())}
                    trace(level, _to_unicode(text).format(**kwargs_unicode))
                except(KeyError, ValueError, UnicodeEncodeError):
                    raise_exception(max(VERBOSE, level + 1))
                    sys.stderr.write("Warning: Problem in trace_fmtd: {exc}\n".
                                     format(exc=sys.exc_info()))
                    # Show arguments so trace contents recoverable
                    sys.stderr.write("   text=%r\n" % _to_utf8(clip_value(text)))
                    ## OLD: kwargs_spec = ", ".join(("%s:%r" % (k, clip_value(v))) for (k, v) in kwargs.iteritems())
                    kwargs_spec = ", ".join(("%s:%r" % (k, clip_value(v))) for (k, v) in list(kwargs.items()))
                    sys.stderr.write("   kwargs=%s\n" % _to_utf8(kwargs_spec))
            except(AttributeError):
                # Note: This can occur when profile_function set
                raise_exception(max(VERBOSE, level + 1))
                sys.stderr.write("Error: Unexpected problem in trace_fmtd: {exc}\n".
                                 format(exc=sys.exc_info()))
        return


    STANDARD_TYPES = (int, float, dict, list)
    SIMPLE_TYPES = (bool, int, float, type(None), str)
    #
    def trace_object(level, obj, label=None, show_all=None, show_private=None, show_methods_etc=None, indentation=None, pretty_print=None, max_value_len=max_trace_value_len, max_depth=0, regular_standard=False):
        """Trace out OBJ's members to stderr if at trace LEVEL or higher.
        Note: Optionally uses output LABEL, with INDENTATION, SHOWing_ALL members, and PRETTY_PRINTing.
        TODO: Use SHOW_PRIVATE to display private members and SHOW_METHODS_ETC for methods.
        Unless REGULAR_STANDARD, object like lists and dicts treated specially.
        If MAX_DEPTH > 0, this uses recursion to show values for instance members."""
        # HACK: Members for STANDARD_TYPES omitted unless show_all.
        # TODO: Make REGULAR_STANDARD True by default
        # Notes:
        # - This is intended for arbitrary objects, use trace_values for objects known to be lists or hashes.
        # - Support for show_private and show_methods_etc is not yet implemented (added for sake of tpo_common.py).
        # - See https://stackoverflow.com/questions/383944/what-is-a-python-equivalent-of-phps-var-dump.
        # TODO: support recursive trace; specialize show_all into show_private and show_methods
        # TODO: handle tuples
        ##                                       r=(trace_level < level)))
        trace_fmt(MOST_VERBOSE, "trace_object({dl}, {obj}, label={lbl}, show_all={sa}, indent={ind}, pretty={pp}, max_d={md})",
                  dl=level, obj=object, lbl=label, sa=show_all, ind=indentation, pp=pretty_print, md=max_depth)
        if (trace_level < level):
            return
        if (pretty_print is None):
            pretty_print = (trace_level > level)
        type_id_label = str(type(obj)) + " " + hex(id(obj))
        if label is None:
            ## BAD: label = str(type(obj)) + " " + hex(hash(obj))
            ## OLD: label = str(type(obj)) + " " + hex(id(obj))
            label = type_id_label
        elif verbose_debugging():
            label += " [" + type_id_label + "]"
        else:
            pass
        outer_indentation = ""
        if indentation is None:
            indentation = INDENT1
        else:
            # TODO: rework via indent_level arg
            outer_indentation = indentation[:len(indentation)-len(INDENT)]
        if show_all is None:
            show_all = (show_private or show_methods_etc)
        if para_mode_tracing:
            trace(ALWAYS, "")
        trace(ALWAYS, outer_indentation + label + ": {")
        ## OLD: for (member, value) in inspect.getmembers(obj):
        member_info = []
        try:
            member_info = inspect.getmembers(obj)
        except:
            trace_fmtd(QUITE_VERBOSE, "Warning: Problem getting member list in trace_object: {exc}",
                       exc=sys.exc_info())
        ## HACK: show standard type value as special member
        if (isinstance(obj, STANDARD_TYPES) and (not regular_standard)):
            member_info = [("(value)", obj)] + [(("__(" + m + ")__"), v) for (m, v) in member_info]
            trace_fmtd(QUITE_VERBOSE, "{ind}Special casing standard type as member {m}",
                       ind=indentation, m=member_info[0][0])
        for (member, value) in member_info:
            # If high trace level, output the value as is
            # TODO: value = clip_text(value)
            trace_fmtd(MOST_DETAILED, "{i}{m}={v}; type={t}", i=indentation, m=member, v=value, t=type(value))
            value_spec = format_value(value, max_len=max_value_len)
            if (trace_level >= MOST_VERBOSE):
                do_print(indentation + member + ": ", end="")
                if pretty_print:
                    pprint(value_spec, stream=sys.stderr)
                    if debug_file:
                        pprint(value_spec, stream=debug_file)
                else:
                    do_print(value_spec)
                if use_logging:
                    logging.debug(_to_utf8((indentation + member + ": " + value_spec)))
                continue
            # Include unless special member (or if no filtering)
            ## DEBUG: trace_expr(QUITE_VERBOSE, member.startswith("__"), re.search(r"^<.*(method|module|function).*>$", value_spec))
            include_member = (show_all or (not (member.startswith("__") or 
                                                re.search(r"^<.*(method|module|function).*>$", value_spec))))
            # Optionally, process recursively (TODO: make INDENT an env. option)
            is_simple_type = isinstance(value, SIMPLE_TYPES)
            if ((max_depth > 0) and include_member and (not is_simple_type)):
                # TODO: add helper for formatting type & address (for use here and above)
                member_type_id_label = (member + " [" + str(type(value)) + " " + hex(id(value)) + "]")
                trace_object(level, value, label=member_type_id_label, show_all=show_all,
                             indentation=(indentation + INDENT), pretty_print=None,
                             max_depth=(max_depth - 1), max_value_len=max_value_len,
                             regular_standard=regular_standard)
                continue
            # Otherwise, derive value spec. (trapping for various exceptions)
            ## TODO: pprint.pprint(member, stream=sys.stderr, indent=4, width=512)
            try:
                try:
                    ## OLD: value_spec = format_value("%r" % ((value),), max_len=max_value_len)
                    if is_simple_type and not isinstance(value, str):
                        value_spec = value
                    else:
                        value_spec = format_value("%r" % ((value),), max_len=max_value_len)
                except(TypeError, ValueError):
                    trace_fmtd(QUITE_VERBOSE, "Warning: Problem in tracing member {m}: {exc}",
                               m=member, exc=sys.exc_info())
                    value_spec = "__n/a__"
            except(AttributeError):
                # Note: This can occur when profile_function set
                trace_fmtd(QUITE_VERBOSE, "Error: unexpected problem in trace_object: {exc}",
                           exc=sys.exc_info())
                value_spec = "__n/a__"
            if include_member:
                do_print(indentation + member + ": ", end="")
                if pretty_print:
                    # TODO: remove quotes from numbers and booleans
                    pprint(value_spec, stream=sys.stderr, indent=len(indentation))
                else:
                    do_print(_to_utf8(value_spec))
                if use_logging:
                    logging.debug(_to_utf8((indentation + member + ":" + value_spec)))
        trace(ALWAYS, indentation + "}")
        if para_mode_tracing:
            trace(ALWAYS, "")
        return


    def trace_values(level, collection, label=None, indentation=None, use_repr=None, max_len=None):
        """Trace out elements of array or hash COLLECTION if at trace LEVEL or higher"""
        trace_fmt(MOST_VERBOSE, "trace_values(dl, {coll}, label={lbl}, indent={ind})",
                  dl=level, lbl=label, coll=collection, ind=indentation)
        if (trace_level < level):
            return
        if para_mode_tracing:
            trace(ALWAYS, "")
        if not isinstance(collection, (list, dict)):
            if hasattr(collection, '__iter__'):
                trace(level + 1, "Warning: [trace_values] consuming iterator")
                collection = list(collection)
            else:
                trace(level + 1, "Warning: [trace_values] coercing input into list")
                collection = [collection]
        if indentation is None:
            indentation = INDENT1
        if label is None:
            ## BAD: label = str(type(collection)) + " " + hex(hash(collection))
            label = str(type(collection)) + " " + hex(id(collection))
            ## OLD: indentation = INDENT1
        if use_repr is None:
            use_repr = False
        trace(ALWAYS, label + ": {")
        keys_iter = list(collection.keys()) if isinstance(collection, dict) else range(len(collection))
        for k in keys_iter:
            try:
                value = format_value(_to_utf8(collection[k]))
                if use_repr:
                    value = repr(value)
                trace_fmtd(ALWAYS, "{ind}{k}: {v}", ind=indentation, k=k,
                           v=format_value(value, max_len=max_len))
            except:
                trace_fmtd(QUITE_VERBOSE, "Warning: Problem tracing item {k}",
                           k=_to_utf8(k), exc=sys.exc_info())
        trace(ALWAYS, indentation + "}")
        if para_mode_tracing:
            trace(ALWAYS, "")
        return


    def trace_expr(level, *values, **kwargs):
        """Trace each of the arguments (if at trace LEVEL or higher), using introspection
        to derive label for each expression. By default, the following format is used:
           expr1=value1; ... exprN=valueN
        Notes:
        - Warning: introspection fails to resolve expressions if statement split across lines.
        - For simplicity, the values are assumed to separated by ', ' (or expression _SEP)--barebones parsing applied.
        - Use DELIM to specify delimiter; otherwise '; ' used;
          if so, NO_EOL applies to intermediate values (EOL always used at end).
        - Use USE_REPR=False to use tracing via str instead of repr.
        - Use _KW_ARG for KW_ARG in case of conflict, as in following:
          trace_expr(DETAILED, term, _term="; ")
        - Use MAX_LEN to specify maximum value length.
        - Use PREFIX to specify initial trace output
        - See misc_utils.trace_named_objects for similar function taking string input, which is more general but harder to use and maintain"""
        trace_fmt(MOST_VERBOSE, "trace_expr({l}, a={args}, kw={kw}); debug_level={dl}",
                  l=level, args=values, kw=kwargs, dl=trace_level)
        ## DEBUG:
        ## trace_fmt(1, "(global_trace_level:{g} < level:{l})={v}",
        ##           g=trace_level, l=level, v=(trace_level < level))
        if (trace_level < level):
            # note: Short-circuits processing to avoid errors about known problem (e.g., under ipython)
            return
        # Note: checks alternative keyword first, so False ones not misintepretted
        sep = kwargs.get('_sep') or kwargs.get('sep')
        delim = kwargs.get('_delim') or kwargs.get('delim')
        no_eol = kwargs.get('_no_eol') or kwargs.get('no_eol')
        in_no_eol = no_eol
        use_repr = kwargs.get('_use_repr') or kwargs.get('use_repr')
        max_len = kwargs.get('_max_len') or kwargs.get('max_len')
        prefix = kwargs.get('_prefix') or kwargs.get('prefix')
        if sep is None:
            sep = ", "
        if no_eol is None:
            no_eol = (delim and ("\n" in delim))
        if delim is None:
            delim = "; "
            if in_no_eol is None:
                no_eol = True
        if use_repr is None:
            use_repr = True
        if prefix is None:
            prefix = ""
        trace(9, f"sep={sep!r}, del={delim!r}, noe={no_eol}, rep={use_repr}, len={max_len}, pre={prefix}")
        # Get symbolic expressions for the values
        # TODO: handle cases split across lines
        try:
            # TODO: rework introspection following icecream (e.g., using abstract syntax tree)
            caller = inspect.stack()[1]
            ## OLD: (_frame, filename, line_number, _function, _context, _index) = caller
            (_frame, filename, line_number, _function, context, _index) = caller
            trace(9, f"filename={filename!r}, context={context!r}")
            statement = read_line(filename, line_number).strip()
            if statement == MISSING_LINE:
                ## OLD: statement = str(context).replace(")\\n']", "")
                statement = str(context).replace("\\n']", "")
            # Extract list of argument expressions (removing optional comment)
            statement = re.sub(r"#.*$", "", statement)
            statement = re.sub(r"^\s*\S*trace_expr\s*\(", "", statement)
            # Remove trailing paren with optional semicolon
            statement = re.sub(r"\)\s*;?\s*$", "", statement)
            # Remove trailing comma (e.g., if split across lines)
            statement = re.sub(r",?\s*$", "", statement)
            # Skip first argument (level)
            ## BAD: expressions = statement.split(sep)[1:]
            expressions = re.split(", +", statement)[1:]
            trace(9, f"expressions={expressions!r}\nvalues={values!r}")
        except:
            trace_fmtd(ALWAYS, "Exception isolating expression in trace_vals: {exc}",
                       exc=sys.exc_info())
            expressions = []
        trace(level, prefix, no_eol=no_eol)
        for expression, value in zip_longest(expressions, values):
            try:
                # Exclude kwarg params
                match = re.search(r"^(\w+)=", str(expression))
                if (match and match.group(1) in kwargs):
                    continue
                assertion((not ((value is not None) and (expression is None))),
                          "Warning: Likely problem resolving expression text (try reworking trace_expr call)")
                value_spec = format_value(repr(value) if use_repr else value,
                                          max_len=max_len)
                trace(level, f"{expression}={value_spec}{delim}", no_eol=no_eol)
            except:
                trace_fmtd(ALWAYS, "Exception tracing values in trace_vals: {exc}",
                       exc=sys.exc_info())            
        ## OLD: if no_eol:
        if (no_eol and (delim != "\n")):
            trace(level, "", no_eol=False)
        return

    
    def trace_current_context(level=QUITE_DETAILED, label=None, 
                              show_methods_etc=False):
        """Traces out current context (local and global variables), with output
        prefixed by "LABEL context" (e.g., "current context: {\nglobals: ...}").
        Notes: By default the debugging level must be quite-detailed (6).
        If the debugging level is higher, the entire stack frame is traced.
        Also, methods are omitted by default."""
        frame = None
        if label is None:
            label = "current"
        try:
            frame = inspect.currentframe().f_back
        except (AttributeError, KeyError, ValueError):
            trace_fmt(VERBOSE, "Exception during trace_current_context: {exc}",
                      exc=sys.exc_info())
        if para_mode_tracing:
            trace(level, "")
        trace_fmt(level, "{label} context: {{", label=label)
        prefix = INDENT
        if (get_level() - level) > 1:
            trace_object((level + 2), frame, "frame", indentation=prefix,
                         show_all=show_methods_etc)
        else:
            trace_fmt(level, "frame = {f}", f=frame)
            if frame:
                trace_object(level, frame.f_globals, "globals", indentation=prefix,
                             show_all=show_methods_etc)
                trace_object(level, frame.f_locals, "locals", indentation=prefix,
                             show_all=show_methods_etc)
        trace(level, "}")
        if para_mode_tracing:
            trace(level, "")
        return


    ## TODO
    ## def trace_stack(level=VERBOSE):
    ##     """Output stack trace to stderr (if at trace LEVEL or higher)"""
    ##     system-ish.print_full_stack()
    ##     return


    def trace_exception(level, task):
        """Trace exception information regarding TASK (e.g., function) at LEVEL"""
        # Note: Conditional output version of system's print_exception_info.
        # ex: trace_exception(DETAILED, "tally_counts")
        trace_fmt(level, "Exception during {t}: {exc}".
                  format(t=task, exc=sys.exc_info()))
        # TODO: include full stack trace (e.g., at LEVEL + 2)
        ## system.trace_stack(level + 2)
        return

    
    def raise_exception(level=1):
        """Raise an exception if debugging (at specified trace LEVEL)"""
        # Note: For producing full stacktrace in except clause when debugging.
        # TODO: elaborate
        if __debug__ and (level <= trace_level):
            raise                       # pylint: disable=misplaced-bare-raise
        return


    def assertion(expression, message=None, assert_level=None):
        """Issue warning if EXPRESSION doesn't hold, along with optional MESSAGE
        Note:
        - Warning: introspection fails to resolve expression if split across lines.
        - This is a "soft assertion" that doesn't raise an exception (n.b., provided the test doesn't do so).
        - The optional ASSERT_LEVEL overrides use of ALWAYS.
        - Returns expression text or None if not triggered.
        """
        # EX: assertion((2 + 2) != 5)
        # TODO: have streamlined version using sys.write that can be used for trace and trace_fmtd sanity checks about {}'s
        # TODO: trace out local and globals to aid in diagnosing assertion failures; ex: add automatic tarcing of variables used in the assertion expression)
        expression_text = None
        if (assert_level is None):
            assert_level = ALWAYS
        if (trace_level < assert_level):
            # note: Short-circuits processing to avoid extraneous warnings (e.g., trace_expr under ipython)
            return expression_text
        if (not expression):
            try:
                # Get source information for failed assertion
                trace_fmtd(MOST_VERBOSE, "Call stack: {st}", st=inspect.stack())
                caller = inspect.stack()[1]
                ## OLD: (_frame, filename, line_number, _function, _context, _index) = caller
                (_frame, filename, line_number, _function, context, _index) = caller
                trace(8, f"filename={filename!r}, context={context!r}")
                # Read statement in file and extract assertion expression
                # TODO: handle #'s in statement proper (e.g., assertion("#" in text))
                statement = read_line(filename, line_number).strip()
                if statement == MISSING_LINE:
                    ## OLD: statement = str(context).replace(")\\n']", "")
                    statement = str(context).replace("\\n']", "")
                # Format expression and message
                # note: removes comments, along with the assertion call prefix and suffix
                statement = re.sub("#.*$", "", statement)
                statement = re.sub(r"^(\S*)assertion\(", "", statement)
                expression = re.sub(r"\);?\s*$", "", statement)
                expression_text = expression
                qualification_spec = (": " + message) if message else ""
                # Output information
                # TODO: omit subsequent warnings
                trace_fmtd(ALWAYS, "Assertion failed: {expr} (at {file}:{line}){qual}",
                           expr=expression, file=filename, line=line_number, qual=qualification_spec)
            except:
                trace_fmtd(ALWAYS, "Exception formatting assertion: {exc}",
                           exc=sys.exc_info())
                trace_object(ALWAYS, inspect.currentframe(), "caller frame", pretty_print=True)
        return expression_text

    def val(level, value):
        """Returns VALUE if at trace LEVEL or higher otherwise None
        Note: inspired by Lisp's convenient IF form without an explicit else: (if test value-if-true)"""
        # EX: (101 if ((get_level() == 1) and val(1, 101)) else None) => 101
        # EX: ((not __debug__) and val(trace_level, 101))) => None
        # TODO: rename as cond_value???
        return (value if (trace_level >= level) else None)

    def code(level, no_arg_function):
        """Execute NO_ARG_FUNCTION if at trace LEVEL or higher
        Notes:
        - Given the quirks of Python syntax, a two-step process is required:
           debug.code(4, { line1; line2; ...; lineN })
               =>
           def my_stupid_block_workaround(): 
               if __debug__:
                   line1; line2; ...; lineN
           debug.code(4, my_stupid_block_workaround)
        - Lambda functions can be used for simple expression-based functions"""
        trace_object(VERBOSE, f"code({level}, {no_arg_function})")
        if (trace_level >= level):
            trace_object(QUITE_DETAILED, f"Executing {no_arg_function}")
            no_arg_function()
        return

else:

    trace_level = 0
    
    def non_debug_stub(*_args, **_kwargs):
        """Non-debug stub (i.e., no-op function)"""
        # Note: no return value assumed by debug.expr
        return


    def get_level():
        """Returns tracing level (i.e., 0)"""
        return trace_level


    def get_output_timestamps():
        """Non-debug stub"""
        return False


    set_output_timestamps = non_debug_stub

    trace = non_debug_stub

    trace_fmtd = non_debug_stub

    trace_object = non_debug_stub

    trace_values = non_debug_stub
    
    trace_expr = non_debug_stub
    
    trace_current_context = non_debug_stub

    trace_exception = non_debug_stub
    
    raise_exception = non_debug_stub
    
    assertion = non_debug_stub

    code = non_debug_stub

    ## TODO?:
    ## val = non_debug_stub
    ##
    def val(_expression):
        """Non-debug stub for value()--a no-op function"""
        # Note: implemented separately from non_debug_stub to ensure no return value
        return
    ##
    ## OLD: assert val(1) is None, "non-debug val() should not return a non-Null value"
    if val(1) is None:
        system.print_error("Warning: non-debug val() should return Null")

# Aliases for terse functions
    
cond_code = code

cond_val = val
    
# note: adding alias for trace_fmtd to account for common typo
# TODO: alias trace to trace_fmt as well (add something like trace_out if format not desired)
trace_fmt = trace_fmtd

def debug_print(text, level):
    """Wrapper around trace() for backward compatibility
    Note: debug_print will soon be deprecated."""
    return trace(level, text)


def timestamp():
    """Return timestamp for use in logging, etc."""
    return (str(datetime.now()))
    

## OLD: def debugging(level=ERROR):
def debugging(level=USUAL):
    """Whether debugging at specified trace LEVEL (e.g., 3 for usual)"""
    ## BAD: """Whether debugging at specified trace level, which defaults to {l}""".format(l=ERROR)
    ## NOTE: Gotta hate python/pylint (no warning about docstring)
    ## TODO: use level=WARNING (i.e., 2)
    return (get_level() >= level)


def detailed_debugging():
    """Whether debugging with trace level DETAILED (4) or higher"""
    ## BAD: """Whether debugging with trace level at or above {l}""".format(l=DETAILED)
    return (get_level() >= DETAILED)


def verbose_debugging():
    """Whether debugging with trace level VERBOSE (5) or higher"""
    ## BAD: """Whether debugging with trace level at or above {l}""".format(l=VERBOSE)
    return (get_level() >= VERBOSE)


def _getenv_bool(name, default_value):
    """Version of system.getenv_bool w/o tracing"""
    ## EX: os.setenv("FU", "1"); _getenv_bool("FU", False)) => True
    result = default_value
    if (str(os.environ.get(name) or default_value).upper() in ["1", "TRUE"]):
        result = True
    return result


def _getenv_int(name, default_value):
    """Version of system.getenv_int w/o tracing"""
    result = default_value
    try:
        env_value = os.environ.get(name)
        if env_value is not None:
            result = int(env_value)
    except:
        _print_exception_info("_getenv_int")
    return result


def format_value(value, max_len=None, strict=None):
    """Format VALUE for output with trace_values, etc.: truncates if too long and encodes newlines
    Note: With STRICT, MAX_LEN is maximum length for returned string (i.e., including "...") unless OLD_MAX_SPEC
    """
    # EX: format_value("    \n\n\n\n", max_len=11) => "    \\n\\n..."
    # EX: format_value("fubar", max_len=3) => "fub..."
    # EX: format_value("fubar", max_len=3, strict=True) => "..."
    # TODO2: rework with result determined via repr
    trace(1 + MOST_VERBOSE, f"format_value({value!r}, max_len={max_len})")
    if max_len is None:
        max_len = max_trace_value_len
    if strict is None:
        strict = False
    result = value if isinstance(value, str) else str(value)
    result = re.sub("\n", r"\\n", result)
    ellipsis = "..."
    extra = (len(result) - max_len)
    if (extra <= 0):
        pass
    elif not strict:
        result = result[:-extra] + ellipsis
    else:
        l = 2 + MOST_VERBOSE
        trace(l, f"0. {result!r}")
        extra2 = 0
        if (len(result) - extra + len(ellipsis) > max_len):
            extra2 = (len(result) - extra + len(ellipsis) - max_len)
        trace_expr(l, extra, extra2)
        result = result[:-(extra + extra2)]
        trace(l, f"1. {result!r}")
        result += ellipsis
        trace(l, f"2. {result!r}")
        result = result[:max_len]
        trace(l, f"3. {result!r}")
        assertion(len(result) <= max_len)
    trace(MOST_VERBOSE, f"format_value() => {result!r}")
    return result


def xor(value1, value2):
    """Whether VALUE1 and VALUE2 differ when interpretted as booleans"""
    # Note: Used to clarify assertions; same as bool(value1) != bool(value2).
    # See https://stackoverflow.com/questions/432842/how-do-you-get-the-logical-xor-of-two-variables-in-python
    # EX: not(xor(0, 0.0))
    result = (bool(value1) ^ bool(value2))
    trace_fmt(QUITE_VERBOSE, "xor({v1}, {v2}) => {r}", v1=value1, v2=value2, r=result)
    return result


def xor3(value1, value2, value3):
    """Whether one and only one of VALUE1, VALUE2, and VALUE3 are true"""
    ## result = (xor(value1, xor(value2m value3))
    ##           and not (bool(value1) and bool(value2) and bool(value3)))
    num_true = sum(int(bool(v)) for v in [value1, value2, value3])
    result = (num_true == 1)
    trace_fmt(QUITE_VERBOSE, "xor3({v1}, {v2}, {v3}) => {r}",
              v1=value1, v2=value2, v3=value3, r=result)
    return result


def init_logging():
    """Enable logging with INFO level by default or with DEBUG if detailed debugging"""
    trace(DETAILED, "init_logging()")
    trace_object(QUITE_DETAILED, logging.root, "logging.root")

    # Set the level for the current module
    # TODO: use mapping from symbolic LEVEL user option (e.g., via getenv)
    level = logging.DEBUG if detailed_debugging() else logging.INFO
    trace_fmt(VERBOSE, "Setting logger level to {ll}", ll=level)
    logging.basicConfig(format='%(asctime)s : %(levelname)s : %(message)s', level=level)
    logging.debug("init_logging()")

    # Optionally make sure logging level applied globally
    if _getenv_bool("GLOBAL_LOGGING", False):
        old_level = logging.root.level
        trace_fmt(VERBOSE, "Setting root logger level from {ol} to {nl}", ol=old_level, nl=level)
        logging.root.setLevel(level)
    return


def _print_exception_info(task):
    """Output exception information to stderr regarding TASK (e.g., function)"""
    # Note: non-tracing version of system's print_exception_info
    sys.stderr.write("Error during {t}: {exc}\n".
                     format(t=task, exc=sys.exc_info()))
    return


def profile_function(frame, event, arg):
    """Function for monitoring function entry and exit (FEE), etc., currently just tracing at level 4. See sys.setprofile.
    # Note: Use a package like viztracer for non-trivial monitoring"""
    # Based on dependency-checking code in PyTorch. See
    #   https://fossies.org/linux/pytorch/torch/package/analyze/trace_dependencies.py.
    # This tries to determine the name of the module that the function is in:
    # 1) Check the global namespace of the frame.
    # 2) Check the local namespace of the frame.
    # 3) To handle class instance method calls, check the attribute named 'name'
    #    of the object in the local namespace corresponding to "self".
    # TODO:
    # - Make tracing level an option (e.g., environment).
    trace_fmt(QUITE_DETAILED, "profile_function(_, {e}, {a})", e=event, a=arg)
    trace_object(QUITE_VERBOSE, frame, "frame")

    # Resolve the names for the function (callable) and module
    name = "???"
    module = "???"
    try:
        name = frame.f_code.co_name
        if name in frame.f_globals:
            module = frame.f_globals[name].__module__
        elif name in frame.f_locals:
            module = frame.f_locals[name].__module__
        elif "self" in frame.f_locals:
            method = getattr(frame.f_locals["self"], name, None)
            module = method.__module__ if method else None
        else:
            trace(QUITE_VERBOSE, "Unable to resolve module")
    except:
        _print_exception_info("profile_function")

    # Trace out the call (with other information at higher tracing levels)
    if event.endswith("call"):
        trace_fmt(DETAILED, "in: {mod}:{func}({a})", mod=module, func=name, a=arg)
    elif event.endswith("return"):
        trace_fmt(DETAILED, "out: {mod}:{func} => {a}", mod=module, func=name, a=arg)
    else:
        trace_fmt(VERBOSE, "profile {mod}:{func} {e}: arg={a}",
                  mod=module, func=name, e=event, a=arg)
    return

def reference_var(*args):
    """No-op function used for referencing variables in ARGS"""
    trace(MOST_VERBOSE, f"reference_var{tuple(args)}")
    return

#-------------------------------------------------------------------------------
# Utility functions useful for debugging (e.g., for trace output)

# TODO: CLIPPED_MAX = system-ish.getenv_int("CLIPPED_MAX", 132)
CLIPPED_MAX = 132
#
def clip_value(value, max_len=CLIPPED_MAX):
    """Return clipped version of VALUE (e.g., first MAX_LEN chars)"""
    # TODO3: replace with format_value
    # TODO: omit conversion to text if already text [DUH!]
    clipped = "%s" % value
    if (len(clipped) > max_len):
        clipped = clipped[:max_len] + "..."
    return clipped

def read_line(filename, line_number):
    """Returns contents of FILENAME at LINE_NUMBER
    Note: returns '???' upon exception
    """
    # ex: "debugging" in read_line(os.path.join(os.getcwd(), "debug.py"), 3)
    # TODO: use rare Unicode value instead of "???"
    try:
        file_handle = open(filename, encoding="UTF-8")
        line_contents = (list(file_handle))[line_number - 1]
        file_handle.close()
    except:
        line_contents = MISSING_LINE
    return line_contents

#-------------------------------------------------------------------------------

def main(args):
    """Supporting/ code for command-line processing"""
    trace_expr(DETAILED, len(args))
    trace(ERROR, "FYI: Not intended for direct invocation. Some tracing examples follow.")
    #
    trace(ALWAYS, "date record for now at trace level 1")
    trace_object(ERROR, datetime.now(), label="now")
    trace(DETAILED, "stack record with max depth 1")
    trace_object(DETAILED, inspect.stack(), label="stack", max_depth=1)
    #
    # Make sure trace_expr traces at proper tracing level
    trace(ALWAYS, "level=N         for N from 0..trace_level ({l})".
          format(l=trace_level))
    for level in range(trace_level):
        level_value = level
        trace_expr(level, level_value)
    #
    # Maker sure trace_expr gets all arguments
    trace(ALWAYS, "n-i=N-i         for N=trace_level ({l}) and i from 1..-1".format(l=trace_level))
    n = trace_level
    i = 1
    ## TODO: for i in range(3, 0, -1):
    trace_expr(ERROR, n+i, n, n+i)
    return

# Do debug-only processing (n.b., for when PYTHONOPTIMIZE not set)
# Note: wrapped in function to keep things clean

if __debug__:

    def open_debug_file():
        """Open external file for copy of trace output"""
        trace(5, "open_debug_file()")
        global debug_file
        assertion(debug_file is None)

        # Open the file
        debug_filename = os.getenv("DEBUG_FILE")
        if debug_filename is not None:
            ## OLD: debug_file = open(debug_filename, mode="w", encoding="UTF-8")
            ## TEST: open unbuffered which requires binary output mode
            ## BAD: debug_file = open(debug_filename, mode="wb", buffering=0, encoding="UTF-8")
            ## note: uses line buffering
            ## OLD: for_append = _getenv_bool("DEBUG_FILE_APPEND", False) or _getenv_bool("DEBUG_FILE_HACK", False)
            for_append = _getenv_bool("DEBUG_FILE_APPEND", True)
            mode = ("a" if for_append else "w")
            trace_expr(5, mode)
            debug_file = open(debug_filename, mode=mode, buffering=1, encoding="UTF-8")
        trace_fmtd(VERBOSE, "debug_filename={fn} debug_file={f}",
                   fn=debug_filename, f=debug_file)
        return

    def reopen_debug_file():
        """Re-open debug file to work around concurrent access issues
        Note: The debug file is mainly used with pytest to work around stderr tracing issues"""
        trace(5, "reopen_debug_file()")
        global debug_file
        assertion(debug_file is not None)

        # Close file if opened
        if debug_file:
            debug_file.close()
            debug_file = None

        # Open fresh
        open_debug_file()
        return
                        

    def debug_init():
        """Debug-only initialization"""
        time_start = time.time()
        trace(DETAILED, f"in debug_init(); {timestamp()}")
        # note: shows command invocation unless invoked via "python -c ..."
        command_line = " ".join(sys.argv)
        assertion(command_line)
        if (command_line and (command_line != "-c") and (command_line != "-m")):
            trace(USUAL, command_line)
        trace_expr(DETAILED, sys.argv)
        open_debug_file()

        # Determine whether tracing include time and date
        global output_timestamps
        ## OLD
        ## output_timestamps = (str(os.environ.get("OUTPUT_DEBUG_TIMESTAMPS", False)).upper()
        ##                      in ["1", "TRUE"])
        output_timestamps = _getenv_bool("OUTPUT_DEBUG_TIMESTAMPS", False)
    
        # Show startup time and tracing info
        module_file = __file__
        trace_fmtd(DETAILED, "[{f}] loaded at {t}", f=module_file, t=timestamp())
        trace_fmtd(DETAILED, "trace_level={l}; output_timestamps={ots}", l=trace_level, ots=output_timestamps)

        # Determine other debug-only environment options
        global para_mode_tracing
        para_mode_tracing = _getenv_bool("PARA_MODE_TRACING", para_mode_tracing)
        global max_trace_value_len
        max_trace_value_len = _getenv_int("MAX_TRACE_VALUE_LEN", max_trace_value_len)
        global use_logging
        use_logging = _getenv_bool("USE_LOGGING", use_logging)
        enable_logging = _getenv_bool("ENABLE_LOGGING", use_logging)
        if enable_logging:
            init_logging()
        monitor_functions = _getenv_bool("MONITOR_FUNCTIONS", False)
        if monitor_functions:
            sys.setprofile(profile_function)
        trace_expr(VERBOSE, para_mode_tracing, max_trace_value_len, use_logging, enable_logging, monitor_functions)

        # Show additional information when detailed debugging
        # TODO: sort keys to facilate comparisons of log files
        pre = post = ""
        if para_mode_tracing:
            pre = post = "\n"
        trace_fmt(DETAILED, "{pre}environment: {{\n\t{env}\n}}{post}",
                  env="\n\t".join([(k + ': ' + format_value(os.environ[k]))
                                   for k in sorted(dict(os.environ))]),
                  pre=pre, post=post)

        # Likewise show additional information during verbose debug tracing
        # Note: use debug.trace_current_context() in client module to show module-specific globals like __name__
        trace_expr(MOST_DETAILED, globals(), max_len=65536)

        # Register to show shuttdown time and elapsed seconds
        # TODO: rename to reflect generic-exit nature
        def display_ending_time_etc():
            """Display ending time information"""
            # note: does nothing if stderr closed (e.g., other monitor)
            # TODO: resolve pylint issue with sys.stderr.closed
            trace_object(QUITE_VERBOSE, sys.stderr)
            if sys.stderr.closed:       # pylint: disable=using-constant-test
                return
            elapsed = round(time.time() - time_start, 3)
            trace_fmtd(DETAILED, "[{f}] unloaded at {t}; elapsed={e}s",
                       f=module_file, t=timestamp(), e=elapsed)
            if monitor_functions:
                sys.setprofile(None)
            global debug_file
            if debug_file:
                debug_file.close()
                debug_file = None
        # note: atexit support is enabled by default unless DEBUG_FILE used (n.b., cleanup issues)
        skip_atexit = _getenv_bool("SKIP_ATEXIT", (debug_file is not None))
        trace_expr(4, skip_atexit)
        if not skip_atexit:
            atexit.register(display_ending_time_etc)
        
        return

    
    # Do the initialization
    debug_init()

#-------------------------------------------------------------------------------

trace_expr(MOST_VERBOSE, 888)
#
if __name__ == '__main__':
    main(sys.argv)
else:
    trace_expr(MOST_VERBOSE, 999)



#! /usr/bin/env python
#
# spaccy_ner.py: Runs spaCy for a variety of natural lanuage processing (NLP)
# tasks.
#
# Notes:
# - This is a generalization of spacy_ner.py, which was just for named entity recognition (NER).
# - Unforunately spaCy's document omits important detail that sentiment analysis is not built in!
# - To compensate, sentiment analyzer is based on vader:
#      https://medium.com/swlh/simple-sentiment-analysis-for-nlp-beginners-and-everyone-else-using-vader-and-textblob-728da3dbe33d
# TODO:
# - ** Disable stupid tensorflow warnings (unless feature used): see https://stackoverflow.com/questions/72033928/python-spacy-module-warning-involving-tensorflow-and-libcudart!
# - * Add part-of-speech tagging (see https://spacy.io/api/tagger).
# - Add support for parsing.
# - Integrate state-of-the-art sentiment analysis.
# - Test other sentence splitters besides NLTK and pySBD:
#   ex: Stanford (https://nlp.stanford.edu/software/tokenizer.html).
# - Integrate alternative sentiment analysis modules:
#   ex: https://spacy.io/universe/project/spacy-textblob.
# 

"""Performs various natural lanuage processing (NLP) tasks via spaCy"""

# Standard packages
import re

# Installed packages
import spacy

# Local packages
from mezcla import debug
from mezcla.main import Main, TRACK_PAGES
from mezcla import system
from mezcla import glue_helpers as gh

# Constants (e.g., for arguments)
#
VERBOSE = "verbose"
RUN_NER = "run-ner"
ANALYZE_SENTIMENT = "analyze-sentiment"
## TODO: RUN_POS_TAGGER = "pos-tagging"
## TODO: RUN_TEXT_CATEGORIZER = "text-categorization"
LANG_MODEL = "lang-model"
## TODO: TASK_MODEL = "task-model"
USE_SCI_SPACY = "use-scispacy"
DOWNLOAD_MODEL = "download-model"
SHOW_REPRESENTATION = "show-representation"
## TODO_ARG1 = "TODO-arg1"

# Environment options
COUNT_ENTITIES = system.getenv_bool("COUNT_ENTITIES", False,
                                    "Show counts of entities in verbose output")
SENT_TOKENIZER = system.getenv_text("SENT_TOKENIZER", " ",
                                    "Name of (alternative) sentence tokenizer to use")
USE_PYSBD = system.getenv_bool("USE_PYSBD", (SENT_TOKENIZER.lower() == "pysbd"),
                               "Use pySBD--pragmatic Sentence Boundary Disambiguation")
USE_NLTK = system.getenv_bool("USE_NLTK", (SENT_TOKENIZER.lower() == "nltk"),
                              "Use NLTK--NL Toolkit")

#...............................................................................

# placeholder for optional sentiment module
SentimentIntensityAnalyzer = None

class SentimentAnalyzer(object):
    """Class for analyzing sentiment of sentences or words"""
    # Uses VADER: Valence Aware Dictionary and sEntiment Reasoner
    # Examples:
    # "bad":  {'neg': 1.0, 'neu': 0.0, 'pos': 0.0, 'compound': -0.5423}
    # "good": {'neg': 0.0, 'neu': 0.0, 'pos': 1.0, 'compound': 0.4404}
    analyzer = None
    
    def __init__(self):
        """Class constructor"""
        # Make sure vader is initialized
        global SentimentIntensityAnalyzer
        if not SentimentIntensityAnalyzer:
            debug.trace(2, "dynamic loading of vader")
            # pylint: disable=import-outside-toplevel, import-error
            from vaderSentiment.vaderSentiment import SentimentIntensityAnalyzer as sia
            SentimentIntensityAnalyzer = sia
        self.analyzer = SentimentIntensityAnalyzer()

    def get_score(self, text):
        """Return sentiment score for word(s) in text"""
        debug.trace_fmt(5, "get_sentiment({t})", t=text)
        all_scores = self.analyzer.polarity_scores(text)
        debug.trace_fmt(6, "scores={all}", all=all_scores)
        score = all_scores.get("compound", 0)
        debug.trace_fmt(4, "get_sentiment({t}) => {s}", t=text, s=score)
        return score

#...............................................................................

pysbd = None
## TODO: PySBDFactory = None

APPLY_SPAN_FIX = system.getenv_bool("APPLY_SPAN_FIX", True,
                                    "Apply pySBD contract-mode fix for char_span")

if USE_PYSBD or USE_NLTK:
    # pylint: disable=import-outside-toplevel, import-error
    import pysbd
    ## TODO: from pysbd.utils import PySBDFactory
    debug.assertion(re.search(r"^[3-9]", spacy.__version__), "spaCy 3+ is required for pysbd")

def get_char_span(doc, start, end):
    """Return character span in DOC from START to END or thereabouts (e.g., END - 1)"""
    # Note: with APPLY_SPAN_FIX, the following fix is used:
    #     https://github.com/nipunsadvilkar/pySBD/pull/97
    debug.trace(6, f"get_char_span(_doc, {start}, {end})")
    if APPLY_SPAN_FIX:
        # Note: With alignment mode "contract", the span of all tokens is completely
        # within the character span.
        debug.trace(6, "trying alignment mode contract")
        span = doc.char_span(start, end, alignment_mode="contract")
    else:
        span = doc.char_span(start, end)

    if not span and APPLY_SPAN_FIX:
        # Note: With alignment mode "expand", the span of all tokens is at least
        # partially covered by the character span)
        debug.trace(6, "retrying with alignment mode expand")
        span = doc.char_span(start, end, alignment_mode="expand")
        
    if not span and APPLY_SPAN_FIX:
        # TODO: make this more systematic and also limit to surrounding boundaries
        for (pre, post) in [(0, 1), (1, 0), (0, 2), (1, 2)]:
            try:
                span = doc.char_span(start - pre, end + post, alignment_mode="expand")
            except:
                span = None
            if span:
                debug.trace(6, f"pysbd hack: adjusted span by {(pre, post)} to {(start - pre, end + post)}")
                break
    debug.assertion(span)
    return span

# Note: requires Spacy 3+
@spacy.Language.component('pysbd_sentence_boundaries')
def pysbd_sentence_boundaries(doc):
    """Pragmatic sentence tokenizer component"""
    # Note: based on pySBD/examples/pysbd_as_spacy_component.py
    debug.trace(6, "pysbd_sentence_boundaries(_)")
    seg = pysbd.Segmenter(language="en", clean=False, char_span=True)
    debug.trace_expr(5, seg)
    custom_sent_char_spans = seg.segment(doc.text)
    debug.trace_expr(5, custom_sent_char_spans)

    # Reconcile with spaCy token character spans
    # Note: this runs into problems when word tokenization differs; see
    #    https://github.com/explosion/spaCy/issues/4531
    #    [char_span not returning spans in some cases #4531]
    ## BAD: char_spans = [doc.char_span(sent_span.start, sent_span.end) for sent_span in custom_sent_char_spans]
    #
    spacy_sent_char_spans = [get_char_span(doc, sent_span.start, sent_span.end) for sent_span in custom_sent_char_spans]
    debug.trace_values(5, spacy_sent_char_spans)

    # Set new-sentence flag for token at each starting offset for sentence
    start_token_ids = [span[0].idx for span in spacy_sent_char_spans if span]
    for token in doc:
        token.is_sent_start = (token.idx in start_token_ids)
    debug.assertion(len(custom_sent_char_spans) == sum(token.is_sent_start for token in doc))
    return doc

#...............................................................................

nltk = None

if USE_NLTK:
    # pylint: disable=import-outside-toplevel, import-error
    import nltk
    from pysbd.utils import TextSpan

@spacy.Language.component('nltk_sentence_boundaries')
def nltk_sentence_boundaries(doc):
    """NLTK sentence tokenizer component
       Note: omits the spacing between sentences (unlike spaCy and pySBD)."""
    # Tokenize sentence and make note of starting/ending offsets
    # note: uses TextSpan from pySBD
    debug.trace(6, "nltk_sentence_boundaries(_)")
    sentence_list = nltk.tokenize.sent_tokenize(doc.text)
    debug.trace_expr(5, sentence_list)
    current_end = -1
    custom_sent_char_spans = []
    last_offset = (len(doc.text) - 1)
    ## TODO: debug.assertion(last_offset >= (sum(len(l) for l in sentence_list) - 1)
    for sent in sentence_list:
        new_start = min(1 + current_end, last_offset)
        new_end =  min(new_start + len(sent), last_offset)
        custom_sent_char_spans.append(TextSpan(sent, new_start, new_end))
        current_end = new_end
    debug.trace_expr(5, custom_sent_char_spans)
    
    # Reconcile with spaCy token character spans
    # Note: works around word tokenization differences as in pysbd_sentence_boundaries
    spacy_sent_char_spans = [get_char_span(doc, sent_span.start, sent_span.end) for sent_span in custom_sent_char_spans]
    debug.trace_values(5, spacy_sent_char_spans)
    
    # Set new-sentence flag for token at each starting offset for sentence
    start_token_ids = [span[0].idx for span in spacy_sent_char_spans if span]
    for token in doc:
        token.is_sent_start = (token.idx in start_token_ids)
    debug.assertion(len(sentence_list) == sum(token.is_sent_start for token in doc))
    return doc

#...............................................................................

# Main class
#
class Script(Main):
    """Input processing class"""
    nlp = None
    type_prefix = ":"
    entity_delim = ", "
    entity_quote = '"'
    spacy_model = "en_core_web_lg"
    analyze_sentiment = False
    run_ner = False
    show_representation = False
    verbose = False
    doc = None
    sentiment_analyzer = None
    use_sci_spacy = False
    download_model = False
    show_reprsentation = False
    ## TODO_arg1 = False
    sent_num = 0

    def setup(self):
        """Check results of command line processing"""
        debug.trace_fmtd(5, "Script.setup(): self={s}", s=self)
        self.spacy_model = self.get_parsed_option(LANG_MODEL, self.spacy_model)
        self.analyze_sentiment = self.get_parsed_option(ANALYZE_SENTIMENT, self.analyze_sentiment)
        self.run_ner = self.get_parsed_option(RUN_NER, self.run_ner)
        self.verbose = self.get_parsed_option(VERBOSE, self.verbose)
        default_use_sci_spacy = re.search(r"\bsci\b", self.spacy_model)
        self.use_sci_spacy = self.get_parsed_option(USE_SCI_SPACY, default_use_sci_spacy)
        self.download_model = self.get_parsed_option(DOWNLOAD_MODEL, self.download_model)
        ## self.TODO_arg1 = self.get_parsed_option(TODO_ARG1, self.TODO_arg1)
        do_specific_task = (self.run_ner or self.analyze_sentiment)
        ## TODO: self.show_representation = self.verbose or (not do_specific_task)
        ## TEST: self.show_representation = (not (do_specific_task or TRACK_PAGES))
        default_show_representation = (not (do_specific_task or TRACK_PAGES))
        self.show_representation = self.get_parsed_option(SHOW_REPRESENTATION, default_show_representation)
        self.doc = None

        # Download model from server
        if self.download_model:
            print(f"Downloading spaCy model: {self.spacy_model}")
            gh.run(f"python -m spacy download {self.spacy_model}")
        
        # Load in external module for sentiment analysis (gotta hate spaCy!)
        if self.analyze_sentiment:
            self.sentiment_analyzer = SentimentAnalyzer()

        # Load spaCy language model (normally large model for English)
        debug.trace_fmt(4, "loading spaCy model {m}", m=self.spacy_model)
        try:
            self.nlp = spacy.load(self.spacy_model)
        except:
            system.print_stderr("Problem loading model {m} via spacy: {exc}",
                                m=self.spacy_model, exc=system.get_exception())
            # If model package not properly installed, the model can be
            # created by the following workaround (n.b., uses specific model for clarity);
            #    nlp = spacy.load("en_core_web_sm")
            #        =>
            #    import en_core_web_sm
            #    nlp = en_core_web_sm.load()
            # TODO: Figure out step needed for proper spaCy bookkeeping after 
            # new package is added to python (e.g., en_core_web_sm under site-packages).
            try:
                debug.trace(3, "Warning: Trying eval hack to load model")
                # pylint: disable=import-outside-toplevel, import-error, eval-used
                if self.use_sci_spacy:
                    import scispacy
                    debug.trace_fmt(4, "dir(scispacy): {d}", d=dir(scispacy))
                exec("import " + self.spacy_model)  # pylint: disable=exec-used
                debug.trace_fmt(4, "dir({m}): {d}", m=self.spacy_model, d=eval("dir(" + self.spacy_model + ")"))
                self.nlp = eval(self.spacy_model + ".load()")
            except:
                system.print_stderr("Problem with alternative load of model {m}: {exc}",
                                    m=self.spacy_model, exc=system.get_exception())
        debug.assertion(self.nlp)

        # Disable pipeline components not needed
        unused = ["parser", "tok2vec"]
        if not self.run_ner:
            unused.append("ner")
        for component in unused:
            try:
                self.nlp.disable_pipe(component)
            except:
                system.print_exception_info(f"disable Spacy pipe component {component}")
        debug.trace(4, f"Pipeline components: {[x[0] for x in self.nlp.pipeline]}")

        # Load in optional SpaCy components
        if USE_PYSBD:
            ## TODO: self.nlp.add_pipe(PySBDFactory(self.nlp))
            self.nlp.add_pipe("pysbd_sentence_boundaries", before="parser")
        elif USE_NLTK:
            self.nlp.add_pipe("nltk_sentence_boundaries", before="parser")
        else:
            debug.assertion(SENT_TOKENIZER.lower() == "spacy")
            # Note: senticize is implicitly used when parser enabled
            self.nlp.add_pipe("sentencizer")
                
        # Sanity checks
        ## OLD: debug.assertion(self.nlp)
        debug.assertion(self.type_prefix != self.entity_delim)
                
        debug.trace_object(5, self, label="Script instance")


    def get_entity_spec(self):
        """Return the named entities tagged in the input as string list of typed entities"""
        ## EX: "PERSON:Trump, ORG:the White House"
        ## TODO: return regular list
        debug.trace_fmt(4, "get_entity_spec(): self={s}", s=self)

        # Collect list of entities (e.g., ["PERSON:Elon Musk,", "ORG:SEC"])
        entity_specs = []
        for ent in self.doc.ents:
            ent_text = ent.text
            debug.assertion(self.type_prefix not in ent.label_)
            if (self.type_prefix in ent_text):
                ent_text = (self.entity_quote + ent_text + self.entity_quote)
            entity_specs.append(ent.label_ + self.type_prefix + ent_text)

        # Convert to string
        entity_spec = self.entity_delim.join(entity_specs)
        debug.trace_fmt(5, "get_entity_spec() => {r}", r=entity_spec)
        return entity_spec

    def get_sentiment_score(self, text=None):
        """Returns the overall sentiment score associated with the TEXT (or entire document)"""
        debug.trace_fmt(4, "get_sentiment_score({t}): self={s}", s=self, t=text)
        if text is None:
            text = system.to_string(self.doc)
        # TODO: show word-level sentiment scores, highlighting cases much
        # different from the overall score
        ## BAD: score = self.doc.sentiment
        score = self.sentiment_analyzer.get_score(text)
        debug.trace_fmt(4, "get_sentiment_score() => {r}", r=score)
        return score
    
    def process_line(self, line):
        """Processes current line from input, showing word/lexeme information by default"""
        # TODO: add entity-type filter
        debug.trace_fmtd(6, "Script.process_line({l})", l=line)

        # Analyze the text in the line
        # TODO: allow for embedded sentences
        ## self.doc = self.nlp(re.sub(r"\S", " ", line))
        line = (re.sub(r"\s", " ", line))
        self.doc = self.nlp(line)
        debug.trace_object(7, self.doc, "doc")
        if self.verbose:
            line_text = re.sub(r"\r?\n", " <newline> ", line)
            print(f"input: {line_text}")
        if TRACK_PAGES and self.verbose:
            print("location info (page/sentence):")
        self.sent_num = 0
        for s in self.doc.sents:
            if (str(s) != "\n"):
                self.sent_num += 1
                self.process_sentence(s)
            else:
                debug.trace_fmtd(5, "Ignoring empty sentence (L{n}:O{o})",
                                 n=self.line_num, o=self.char_offset)

    def process_sentence(self, sentence):
        """Process SENTENCE"""
        sentence_text = re.sub(r"[\r\n]$", "", str(sentence), count=1)
        debug.trace_fmtd(6, "Script.process_sentence({s})", s=sentence_text)
        if (sentence_text.strip() == ""):
            debug.trace_fmtd(5, "Ignoring empty sentence")
            return
        sent_info = self.doc[sentence.start: sentence.end]
        if TRACK_PAGES:
            # Notes: Paragraph numbers are relative to each page; and, likewise
            # sentence numbers are relative to each paragraph.
            # See process_line and Main.read_input.
            ## OLD: print(f"Pg#{self.page_num}/Sent#{self.sent_num}: {sentence_text}")
            if self.verbose:
                print(f"Pg#{self.page_num}/Sent#{self.sent_num}: ", end="")
            print(sentence_text)
            debug.trace(4, f"Offset: {self.char_offset}")

        # Show synopsis of word token representations
        # TODO: have options for specifying attributes to show
        if self.show_representation:
            # Gather the word (lexeme) information
            # Note: Based on vocab/lexemes section of https://spacy.io/usage/spacy-101
            word_attributes = ["text", "is_oov", "is_stop", "sentiment"]
            token_attributes = ["is_sent_start"]
            all_info = []
            for word in sent_info:
                lexeme = self.doc.vocab[word.text]
                info = [getattr(lexeme, a, "") for a in word_attributes]
                if (info[0] == "\n"):
                    info[0] = "\\n"
                token_info = [getattr(word, a, "") for a in token_attributes]
                all_info.append("\t".join([system.to_string(v) for v in info + token_info]))
            # Output the synopsis
            prefix = ("\t".join(word_attributes + token_attributes) + "\n")
            doc_repr = "\n".join(all_info)
            print(prefix + doc_repr)

        # Optionally, invoke NER
        if self.run_ner:
            prefix = ""
            if self.verbose:
                if COUNT_ENTITIES:
                    prefix = "{len} ".format(len=len(sent_info.ents))
                prefix += "entities:"
            print(prefix + self.get_entity_spec())
        # Optionally, do sentiment analysis
        if self.analyze_sentiment:
            prefix = "sentiment: " if self.verbose else ""
            sent_text = str(sentence.text)
            print(prefix + str(self.get_sentiment_score(sent_text)))
        
#-------------------------------------------------------------------------------
    
if __name__ == '__main__':
    debug.trace_current_context(level=debug.QUITE_DETAILED)
    app = Script(
        description=__doc__,
        # Note: skip_input controls the line-by-line processing, which is inefficient but simple to
        # understand; in contrast, manual_input controls iterator-based input (the opposite of both).
        skip_input=False,
        manual_input=False,
        # Paragraph mode is required for proper sentence tokenization
        paragraph_mode=True,
        boolean_options=[RUN_NER, ANALYZE_SENTIMENT, VERBOSE,
                         (DOWNLOAD_MODEL, "Download spaCy model"),
                         (SHOW_REPRESENTATION, "Show final representation (e.g., word & token attributes"),
                         ## (TODO_ARG1, "TODO-desc"),
        ],
        text_options=[(LANG_MODEL, "Language model for NLP")])
    app.run()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Test(s) for ../convert_emoticons.py
#

"""Tests for convert_emoticons module"""

# Standard packages
## TODO: from collections import defaultdict

# Installed packages
import pytest

# Local packages
from mezcla.unittest_wrapper import TestWrapper
from mezcla import debug
from mezcla.my_regex import my_re
from mezcla import system

# Note: Two references are used for the module to be tested:
#    THE_MODULE:                  global module object
#    TestTemplate.script_module:  path to file
import mezcla.convert_emoticons as THE_MODULE
#
# Note: sanity test for customization (TODO: remove if desired)
if not my_re.search(__file__, r"\btemplate.py$"):
    debug.assertion("mezcla.template" not in str(THE_MODULE))


D = system.path_separator()

class TestIt(TestWrapper):
    """Class for testcase definition"""
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)
    script_file = my_re.sub(rf"{D}tests{D}test_", f"{D}", __file__)

    @pytest.mark.xfail
    def test_over_script(self):
        """Makes sure works as expected over script itself"""
        debug.trace(4, f"TestIt.test_over_script(); self={self}")
        output = self.run_script(options="", data_file=self.script_file)
        # the example usage should have input emoticon changed to match output
        assert not my_re.search(r"(\[sleeping face\]).*\1", script_contents)
        assert my_re.search(r"(\[sleeping face\]).*\1", output)

        # Make sure no emoticon byte sequences in UTF-8 sequences for script and output
        # note: uses broader UTF8-based tests than Unicdoe character DB used in script
        loose_emoticon_regex = br"[\xE0-\xFF][\x80-\xFF]{1,3}"
        assert (not my_re.search(loose_emoticon_regex, script_contents.encode()))
        assert (my_re.search(loose_emoticon_regex, output.encode()))
        return

    def test_over_script_sans_comments(self):
        """Makes sure works as expected over script itself"""
        debug.trace(4, f"TestIt.test_over_script_sans_comments(); self={self}")
        D = system.path_separator()

        # Strip comments from script and run conversion over it
        script_contents = system.read_file(__file__)
        script_contents = my_re.sub("#.*\n", "\n", script_contents)
        system.write_file(self.temp_file, script_contents)
        output = self.run_script(options="", data_file=self.temp_file)

        # There should be no extended ascii bytes
        assert not my_re.search(b"[\x80-\xFF]", script_contents.encode())
        assert not my_re.search(b"[\x80-\xFF]", output.encode())
        return

class TestIt2:
    """Another class for testcase definition
    Note: Needed to avoid error with pytest due to inheritance with unittest.TestCase via TestWrapper"""
    
    def test_misc(self):
        """Test direct calls for conversion"""
        debug.trace(4, f"TestIt2.test_whatever(); self={self}")
        convert_emoticons = THE_MODULE.convert_emoticons
        cool_smile = "\U0001F60E"        # 😎
        assert (convert_emoticons(cool_smile) == "[smiling face with sunglasses]")
        assert (convert_emoticons(cool_smile, strip=True) == "")
        chinese_age = "\uF9A8"           # 令 ("age" in Chinese)
        assert (convert_emoticons(chinese_age) == chinese_age)         
        return


#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    pytest.main([__file__])



#! /usr/bin/env python
# Example for using the KenLM language modeling utility (based on sample):
#     http://kheafield.com/code/kenlm
# This assumes the language model has already been created via lmplz
#
#--------------------------------------------------------------------------------
# Note: Build steps for C-based utility (based on https://github.com/kpu/kenlm)
# - Requirements:
#   sudo apt install build-essential cmake libboost-system-dev libboost-thread-dev libboost-program-options-dev libboost-test-dev libeigen3-dev zlib1g-dev libbz2-dev liblzma-dev
# - Main code:
#   mkdir -p build
#   cd build
#   cmake ..
#   make -j 4
#   sudo make install
# - Python code:
#   pip install https://github.com/kpu/kenlm/archive/master.zip
# - For more info, see following:
#   https://github.com/kpu/kenlm
#   https://github.com/kpu/kenlm/blob/master/BUILDING
#--------------------------------------------------------------------------------
#
# TODO:
# - Add lint comment disabling extraneous lint warning (or modify python
#   interface to initialize it):
#     Module 'kenlm' has no 'LanguageModel' member (no-member)
#

"""Example for using the KenLM language modeling utility"""

import os
import kenlm
import sys

from mezcla.system import getenv_boolean, getenv_text
from mezcla import glue_helpers as gh
from mezcla import tpo_common as tpo

# Initialize globals, including adhoc options via environment variables
DEFAULT_LM_FILE = os.path.join(os.path.dirname(__file__), '..', 'lm', 'test.arpa')
LM = getenv_text("LM", DEFAULT_LM_FILE)
SENT_DELIM = getenv_text("SENT_DELIM", "\n", "Delimiter for sentence splitting")
VERBOSE = getenv_boolean("VERBOSE", False, "Verbose output")

# Sanity check
show_usage = (len(sys.argv) > 1) and ("--help" in sys.argv[1])
if (not show_usage) and (not gh.non_empty_file(LM)):
    print("Unable to find usable language model file '%s'" % LM)
    show_usage = True
if show_usage:
    print("Usage: %s [--help] [sentence | -]" % sys.argv[0])
    print("")
    print("Example:")
    print("  kenlm=~/programs/kenlm")
    print("  export PATH=$kenlm/bin:$PATH")
    print("  export LM=$kenlm/test.arpa")
    print("  %s" % __file__)
    print("")
    print("Notes:")
    print("- Use LM environment varaiable to specify alternative language model (see lmplz)")
    print("- Use SENT_DELIM to specify setence delimiter (default is newline)")
    print("- To use standard input, specify - for sentence above.")
    print("- Use the following environment options to customize processing")
    print("\t" + tpo.formatted_environment_option_descriptions())
    sys.exit()

# Load model from ARPA-format file
model = kenlm.LanguageModel(LM)
print('{0}-gram model'.format(model.order))

# Define helper function used when checking that total full score equals direct score
def summed_constituent_score(s):
    """Return sum of scores for ngrams for sentence S"""
    ## OLD: return sum(prob for (prob, _) in model.full_scores(s))
    return sum(prob for (prob, _len, _oov) in model.full_scores(s))

# Read input
sentences = 'language modeling is fun .'
if (len(sys.argv) > 1):
    sentences = " ".join(sys.argv[1:])
if (sentences == "-"):
    sentences = sys.stdin.read()

# Check each sentence (or phrase)
# TODO: rework sto avoid reading sentences entirely into memory (e.g., via generic iterator)
for sentence in sentences.split(SENT_DELIM):
    if not sentence.strip():
        continue
    print("sentence: %s" % sentence)
    print("model score: %s" % tpo.round_num(model.score(sentence)))
    normaized_score = (model.score(sentence) / len(sentence.split()))
    print("normalized score: %s" % tpo.round_num(normaized_score))
    ## TODO: gh.assertion(abs(summed_constituent_score(sentence) - model.score(sentence)) < 1e-3)

    # Print diagnostics in verbose mode
    if VERBOSE:
        words = ['<s>'] + sentence.split() + ['</s>']

        # Show scores and n-gram matches
        print("Constituent ngrams")
        ## OLD: print("Offset\tProb\tLength\tWords")
        ## OLD: for i, (prob, length) in enumerate(model.full_scores(sentence)):
        print("Offset\tProb\tLength\tUnseen\tWords")
        for i, (prob, length, unseen) in enumerate(model.full_scores(sentence)):
            start = i + 2 - length
            end = start + length
            ## OLD: print('{0}\t{1}\t{2}\t{3}'.format(i, tpo.round_num(prob), length, ' '.join(words[start : end])))
            unseen_spec = "*" if unseen else ""
            print('{0}\t{1}\t{2}\t{3}\t{4}'.format(i, tpo.round_num(prob), length, unseen_spec, ' '.join(words[start : end])))

        # Find out-of-vocabulary words
        print("out-of-vocabulary words: %s" % [w for w in words if (not w in model)])



#! /usr/bin/env python
#
# Test(s) for ../xml_utils.py
#
# Notes:
# - For debugging the tested script, the ALLOW_SUBCOMMAND_TRACING environment
#   option shows tracing output normally suppressed by  unittest_wrapper.py.
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_xml_utils.py
#
#-------------------------------------------------------------------------------
# Sample input and output:
#
# Input:
#     <?xml version="1.1"?> <xml><a><b>1<c>2<d/>3</c></b>4</a></xml>
#
# XML parse tree
#      xml:
#              a:
#                      b: 1
#                              c: 2
#                                      d:
#                                      3
#                      4
#

"""Tests for xml_utils module"""

# Standard packages
import re

# Installed packages
import pytest

# Local packages
from mezcla.unittest_wrapper import TestWrapper
from mezcla import debug
## TODO: from mezcla import system
from mezcla import glue_helpers as gh

# Note: Two references are used for the module to be tested:
#    THE_MODULE:	    global module object
import mezcla.xml_utils as THE_MODULE
#
# Note: sanity test for customization (TODO: remove if desired)
if not re.search(__file__, r"\btemplate.py$"):
    debug.assertion("mezcla.template" not in str(THE_MODULE))

## TODO:
## # Environment options
## # Note: These are just intended for internal options, not for end users.
## # It also allows for enabling options in one place.
## #
## FUBAR = system.getenv_bool("FUBAR", False,
##                            description="Fouled Up Beyond All Recognition processing")

NESTED_XML = """<?xml version="1.1"?>
<xml><a><b>1<c>2<d/>3</c></b>4</a></xml>
"""

class TestXmlUtils(TestWrapper):
    """Class for testcase definition"""
    script_file = TestWrapper.get_module_file_path(__file__)
    script_module = TestWrapper.get_testing_module_name(__file__)

    def test_data_file(self):
        """Makes sure simple XML data file parsed OK"""
        debug.trace(4, "TestXmlUtils.test_data_file()")
        data = "<xml><a>A</a><b>B</b></xml>"
        gh.write_file(self.temp_file, data)
        output = self.run_script("", self.temp_file)
        assert "a: A" in output
        assert "b: B" in output
        return

    def test_get_xml_text(self):
        """Ensure get_xml_text works as expected"""
        debug.trace(4, "test_get_xml_text()")
        # Example from https://docs.python.org/3/library/xml.etree.elementtree.html.
        parsed_xml_text = THE_MODULE.get_xml_text(THE_MODULE.parse_xml(NESTED_XML))
        assert parsed_xml_text == "xml: \n\ta: \n\t\tb: 1\n\t\t\tc: 2\n\t\t\t\td: \n\t\t\t\t3\n\t\t4"
        return

#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    pytest.main([__file__])



#! /usr/bin/env python
#
# TODO: Test(s) for ../<module>.py
#
# Notes:
# - Fill out TODO's below. Use numbered tests to order (e.g., test_1_usage).
# - TODO: If any of the setup/cleanup methods defined, make sure to invoke base
#   (see examples below for setUp and tearDown).
# - For debugging the tested script, the ALLOW_SUBCOMMAND_TRACING environment
#   option shows tracing output normally suppressed by  unittest_wrapper.py.
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_<module>.py
#

"""TODO: Tests for <module> module"""

# Standard packages
## TODO: from collections import defaultdict

# Installed packages
import pytest

# Local packages
from mezcla.unittest_wrapper import TestWrapper
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla.my_regex import my_re
from mezcla import system

# Note: Two references are used for the module to be tested:
#    THE_MODULE:                  global module object
#    TestTemplate.script_module:  path to file
## TODO (vvv): change template to new name; *** use commented out template below ***
THE_MODULE = None           ## TODO: remove this line (n.b., used just to avoid syntax problems with <module> in following)
## TODO (^^^): import mezcla.<module> as THE_MODULE
#
# Note: sanity test for customization (TODO: remove if desired)
if not my_re.search(__file__, r"\btemplate.py$"):
    debug.assertion("mezcla.template" not in str(THE_MODULE))

## TODO:
## # Environment options
## # Note: These are just intended for internal options, not for end users.
## # It also allows for enabling options in one place.
## #
## FUBAR = system.getenv_bool("FUBAR", False,
##                            description="Fouled Up Beyond All Recognition processing")


class TestIt(TestWrapper):
    """Class for testcase definition"""
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)
    # -or- non-mezcla: script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)
    #
    # TODO: use_temp_base_dir = True            # treat TEMP_BASE as directory
    # note: temp_file defined by parent (along with script_module, temp_base, and test_num)

    ## TODO: optional setup methods
    ##
    ## @classmethod
    ## def setUpClass(cls):
    ##     """One-time initialization (i.e., for entire class)"""
    ##     debug.trace(6, f"TestIt.setUpClass(); cls={cls}")
    ##     # note: should do parent processing first
    ##     super().setUpClass()
    ##     ...
    ##     return
    ##
    ## def setUp(self):
    ##     """Per-test setup"""
    ##     debug.trace(6, f"TestIt.setUp(); self={self}")
    ##     # note: must do parent processing first (e.g., for temp file support)
    ##     super().setUp()
    ##     ...
    ##     # TODO: debug.trace_current_context(level=debug.QUITE_DETAILED)
    ##     return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_data_file(self):
        """Makes sure TODO works as expected"""
        debug.trace(4, f"TestIt.test_data_file(); self={self}")
        data = ["TODO1", "TODO2"]
        system.write_lines(self.temp_file, data)
        ## TODO: add use_stdin=True to following if no file argument
        output = self.run_script(options="", data_file=self.temp_file)
        assert my_re.search(r"TODO-pattern", output.strip())
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_something_else(self):
        """TODO: flesh out test for something else"""
        debug.trace(4, f"TestIt.test_something_else(); self={self}")
        assert False
        ## ex: assert (THE_MODULE.TODO_function() == TODO_value)
        return


    ## TODO: optional cleanup methods
    ##
    ## def tearDown(self):
    ##     debug.trace(6, f"TestIt.tearDown(); self={self}")
    ##     super(TestIt, cls).tearDownClass()
    ##     ...
    ##     return
    ##
    ## @classmethod
    ## def tearDownClass(cls):
    ##     debug.trace(6, f"TestIt.tearDownClass(); cls={cls}")
    ##     super(TestIt, self).tearDown()
    ##     ...
    ##     return

## TODO:
## #...............................................................................
##
## class TestIt2:
##     """Another class for testcase definition
##     Note: Needed to avoid error with pytest due to inheritance with unittest.TestCase via TestWrapper"""
##
##     def test_whatever(self):
##         """TODO: flesh out test for whatever"""
##         debug.trace(4, f"TestIt2.test_whatever(); self={self}")
##         assert False, "TODO: code test"
##         ## ex: assert THE_MODULE.fast_sort() == THE_MODULE.slow_sort()
##         return
##

#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    pytest.main([__file__])



#! /usr/bin/env python
#
# Test(s) for ../bing_search.py
#
# Notes:
# - For debugging the tested script, the ALLOW_SUBCOMMAND_TRACING environment
#   option shows tracing output normally suppressed by  unittest_wrapper.py.
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_bing_search.py
#

"""Tests for bing_search module"""

# Standard packages
## NOTE: this is empty for now

# Installed packages
import pytest

# Local packages
from mezcla import debug

# Note: Two references are used for the module to be tested:
#    THE_MODULE:	    global module object
import mezcla.bing_search as THE_MODULE

class TestBingSearch:
    """Class for testcase definition"""

    ## TODO: TESTS WORK-IN-PROGRESS


if __name__ == '__main__':
    debug.trace_current_context()
    pytest.main([__file__])



#! /usr/bin/env python
#
# Notes:
# - Simple installer for subset of scripts under Github
#       https://github.com/tomasohara/misc-utility
# - Based on following:
#       https://stackoverflow.com/questions/1471994/what-is-setup-py
#

"""Simple installer"""

from distutils.core import setup

setup(name='Mezcla',
      packages=['mezcla'],
      module="mezcla",
      version='1.3.9',
      description-file="README.txt",
      dist-name="Mezcla",
      ## OLD: py_modules=PYTHON_MODULE_NAMES,
      author="Tom O'Hara",
      # TODO: find out which email key is preferred
      email="tomasohara@gmail.com",
      author-email="tomasohara@gmail.com"
      requires-python=">=3.6",
      ## TODO:
      ## install_requires=["six"],
      home-page="https://github.com/tomasohara/Mezcla",
      classifiers=[
          "License :: OSI Approved :: LGPLv3",
          "Programming Language :: Python :: 3",
          "Programming Language :: Python :: 3.7",
      ]
      description="""
Package with core modules from https://github.com/tomasohara/misc-utility
note: mezcla is Spanish for mixture.
""")

[tool.flit.scripts]
realpython = "mezcla.__main__:main"



#! /usr/bin/env python
#
# Produces abstract syntax trees from Bash snippets.
#
# note:
# - Wrapper around bashlex package with AST comparison support added.
#

"""
Support for parsing Bash into abstract syntax trees (ASTs) such as via bash lex

Sample usage:
   {prog} - <<<'sum=$((2 + 2)); if [ $sum -eq 5 ]; then echo "WTH?"; fi'
"""
## TODO: {prog} <<<'if [[ $((2 + 2)) -eq 5 ]]; then echo "WTH?"; fi'


# Standard modules
## TODO

# Installed modules
import bashlex

# Local modules
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla.main import Main
from mezcla.my_regex import my_re
from mezcla import system

# Constants
TL = debug.TL


class BashAST:
    """Abstract syntax tree for Bash snippets
    Example:
    this_ast = BashAST("if true; then echo hey; fi;")
    other_ast = BashAST("echo hey;")
    this_ast.embedded_in(other_ast)
    """
    # note: Usage based on https://github.com/idank/bashlex

    def __init__(self, snippet=None):
        self.snippet = None
        self.parts = None
        if snippet is not None:
            self.parse(snippet)
        return

    def parse(self, snippet):
        """Parse SNIPPET into AST"""
        self.snippet = snippet
        try:
            self.parts = bashlex.parse(snippet)
        except:
            system.print_exception_info("BashAST.parse")
        result = self.parts
        debug.trace(6, f"BashAST.parse({snippet!r}) => {result!r}")
        return result

    def dump(self):
        """Return DUMP of AST"""
        result = ""
        try:
            result = "".join(ast.dump() for ast in self.parts)
        except:
            system.print_exception_info("BashAST.dump")
        return result

    def embedded_in(self, other):
        """Whether AST for current snippet embedded in that for another"""
        # note: implements tree traversal ignoring offsets (i.e., pos attribute)
        return self.ast_embedded_in(self.parts, other.parts)

    @staticmethod
    def ast_embedded_in(ast, other_ast, depth=0):
        """Whether AST is embedded in OTHER (i.e., ignoring offset)"""
        # TODO: ast:List<ast_node_repr>
        indent = (" " * depth)
        debug.trace(8, f"{indent}ast_embedded_in({ast!r}, {other_ast!r}, [d={depth}])")
        embedded = False
        if len(ast) == len(other_ast):
            embedded = all(BashAST.same_ast_node(ast[i], other_ast[i]) for i in range(len(ast)))
        else:
            for sub_node in ast:
                embedded = BashAST.ast_embedded_in(sub_node, other_ast,
                                                   depth=(1 + depth))
                if embedded:
                    break
        if not embedded:
            embedded = (BashAST.ast_node_repr(ast, loose=True)
                        in BashAST.ast_node_repr(other_ast, loose=True))
        debug.trace(7, f"{indent}ast_embedded_in() => {embedded}")
        return embedded
    
    @staticmethod
    def same_ast_node(ast, other_ast):
        """Whether AST node is same as OTHER ignoring offset"""
        # TODO: ast:List<ast_node_repr>
        # example:
        #   CommandNode(pos=(0, 8), parts=[WordNode(pos=(0, 4), word='echo'), WordNode(pos=(5, 8), word='hey'),])
        #   CommandNode(pos=(14, 22), parts=[WordNode(pos=(14, 18), word='echo'), WordNode(pos=(19, 22), word='hey'),])
        same_node = ((ast.kind == other_ast.kind)
                     and (BashAST.ast_node_repr(ast) == BashAST.ast_node_repr(other_ast)))
        debug.trace(7, f"same_ast_node({ast!r}, {other_ast!r}) => {same_node}")
        return same_node

    @staticmethod
    def ast_node_repr(ast, include_pos=False, loose=False):
        """Return representation for node optionally INCLUDing_POS and using LOOSE format"""
        # TODO: ast:List<ast_node_repr>
        node_repr = str(ast)
        if not include_pos:
            node_repr = my_re.sub(r"pos=\(\d+, *\d+\) *", "", node_repr)
        if loose:
            node_repr = my_re.sub(r"^\[(.*)\]$", r"\1", node_repr)
        return node_repr


def main():
    """Entry point"""
    debug.trace(TL.USUAL, f"main(): script={system.real_path(__file__)}")

    # Show simple usage if --help given
    dummy_main_app = Main(description=__doc__.format(prog=gh.basename(__file__)),
                          skip_input=False, manual_input=False)
    debug.assertion(dummy_main_app.parsed_args)

    # Prase Bash snippet from stdin
    snippet = dummy_main_app.read_entire_input()
    ast = BashAST(snippet)
    print(ast.dump())
    return
    
#-------------------------------------------------------------------------------
    
if __name__ == '__main__':
    debug.trace_current_context(level=TL.QUITE_DETAILED)
    main()



#! /usr/bin/env python
#
# Support for performing Term Frequency (TF) Inverse Document Frequency (IDF)
# using ngrams. This is provides a wrapper class around the tfidf package
# by elzilrac (https://github.com/elzilrac/tf-idf).
#
# For details on computations, see following Wikipedia pages:
#    https://en.wikipedia.org/wiki/Tf-idf
#    https://en.wikipedia.org/wiki/N-gram.
#
# Note:
# - This provides the wrapper class ngram_tfidf_analysis around tfidf for use
#   in applications like Visual Diff Search (VDS) that use text from external sources.
# - See compute_tfidf.py for computing tfidf over files.
#
# TODO:
# - Add filtering (e.g., subsumption, all numbers).
# - Isolate ngram support into separate module.
# - Reconcile with compute_tfidf.py (e.g., subsumption here with overlap there).
#

"""TF-IDF using phrasal terms via ngram analysis"""
## TODO: fix description (e.g., add pointer to VDS code)

# Standard packages
## import math
## import os
import re
import sys

# Installed packages
from sklearn.feature_extraction.text import CountVectorizer

# Local packages
from mezcla import debug
from mezcla import system
from mezcla import tpo_common as tpo
from mezcla import tfidf
from mezcla.compute_tfidf import terms_overlap
from mezcla.tfidf.corpus import Corpus as tfidf_corpus
from mezcla.tfidf.preprocess import Preprocessor as tfidf_preprocessor
from mezcla.tfidf import preprocess as tfidf_preprocess

PREPROCESSOR_LANG = system.getenv_text("PREPROCESSOR_LANG", "english",
                                       "Language for ngram preprocessor")
# NOTE: MIN_NGRAM_SIZE (e.g., 2) is alternative to deprecated ALL_NGRAMS (implies 1)
MAX_NGRAM_SIZE = system.getenv_int("MAX_NGRAM_SIZE", 4)
# TODO: add descriptions to all getenv options
MIN_NGRAM_SIZE = system.getenv_int("MIN_NGRAM_SIZE", 2)
ALL_NGRAMS = system.getenv_boolean("ALL_NGRAMS", False)
USE_NGRAM_SMOOTHING = system.getenv_boolean("USE_NGRAM_SMOOTHING", False)
DEFAULT_TF_WEIGHTING = 'basic'
TF_WEIGHTING = system.getenv_text("TF_WEIGHTING", DEFAULT_TF_WEIGHTING)
DEFAULT_IDF_WEIGHTING = 'smooth' if USE_NGRAM_SMOOTHING else 'basic'
IDF_WEIGHTING = system.getenv_text("IDF_WEIGHTING", DEFAULT_IDF_WEIGHTING)
MAX_TERMS = system.getenv_int("MAX_TERMS", 100)
ALLOW_NGRAM_SUBSUMPTION = system.getenv_boolean("ALLOW_NGRAM_SUBSUMPTION", False,
                                                "Allow ngram subsumed by another--substring")
ALLOW_NGRAM_OVERLAP = system.getenv_boolean("ALLOW_NGRAM_OVERLAP", False,
                                            "Allows ngrams to overlap--token boundariese")
ALLOW_NUMERIC_NGRAMS = system.getenv_boolean("ALLOW_NUMERIC_NGRAMS", False)
## OLD: DEFAULT_USE_CORPUS_COUNTER = (not tfidf_preprocessor.USE_SKLEARN_COUNTER)
DEFAULT_USE_CORPUS_COUNTER = (not tfidf_preprocess.USE_SKLEARN_COUNTER)
USE_CORPUS_COUNTER = system.getenv_boolean("USE_CORPUS_COUNTER", DEFAULT_USE_CORPUS_COUNTER,
                                           "Use slow tfidf package ngram tabulation")

try:
    # Note major and minor revision values are assumed to be integral
    major_minor = re.sub(r"^(\d+\.\d+).*", r"\1", tfidf.__version__)
    TFIDF_VERSION = float(major_minor)
except:
    TFIDF_VERSION = 1.0
    system.print_stderr("Exception in main: " + str(sys.exc_info()))
assert(TFIDF_VERSION > 1.0)

            
class ngram_tfidf_analysis(object):
    """Class for performing TF-IDF over ngrams and returning sorted list"""

    def __init__(self, pp_lang=PREPROCESSOR_LANG, min_ngram_size=MIN_NGRAM_SIZE, max_ngram_size=MAX_NGRAM_SIZE, 
                 *args, **kwargs):
        """Class constructor: initialize corpus object (with PP_LANG, MIN_NGRAM_SIZE, and MAX_NGRAM_SIZE)"""
        # TODO: add option for stemmer; add all_ngrams and min_ngram_size to constructor
        debug.trace_fmtd(4, "ngram_tfidf_analysis.__init__(lang={pl}, min={minsz}, max={maxsz})",
                         pl=pp_lang, minsz=min_ngram_size, maxsz=max_ngram_size)
        debug.trace_fmtd(5, "\targs={a} kwargs={k}", a=args, k=kwargs)
        if pp_lang is None:
            pp_lang = PREPROCESSOR_LANG
        self.min_ngram_size = min_ngram_size
        self.max_ngram_size = max_ngram_size
        self.pp = tfidf_preprocessor(language=pp_lang,
                                     gramsize=self.max_ngram_size,
                                     min_ngram_size=self.min_ngram_size,
                                     all_ngrams=ALL_NGRAMS,
                                     stemmer=lambda x: x)
        self.corpus = tfidf_corpus(gramsize=self.max_ngram_size,
                                   min_ngram_size=self.min_ngram_size,
                                   all_ngrams=ALL_NGRAMS,
                                   language=pp_lang,
                                   preprocessor=self.pp)
        super().__init__(*args, **kwargs)

    def add_doc(self, text, doc_id=None):
        """Add document TEXT to collection with key DOC_ID, which defaults to order processed (1-based)"""
        if doc_id is None:
            doc_id = str(len(self.corpus) + 1)
        self.corpus[doc_id] = text

    def get_doc(self, doc_id):
        """Return document data for DOC_ID"""
        return self.corpus[doc_id]

    def get_top_terms(self, doc_id, tf_weight=TF_WEIGHTING, idf_weight=IDF_WEIGHTING, limit=MAX_TERMS,
                      allow_ngram_subsumption=ALLOW_NGRAM_SUBSUMPTION,
                      allow_ngram_overlap=ALLOW_NGRAM_OVERLAP, allow_numeric_ngrams=ALLOW_NUMERIC_NGRAMS):
        """Return list of (term, weight) tuples for DOC_ID up to LIMIT count, using TF_WEIGHT and IDF_WEIGHT schemes
        Notes:
        - TF_WEIGHT in {basic, binary, freq, log, norm_50}
        - IDF_WEIGHT in {basic freq, max, prob, smooth}
        - The top ngrams omit blanks and other relics of tokenization
        - Lower weighted ngrams are omitted if subsumed by higher (or vice versa) unless ALLOW_NGRAM_SUBSUMPTION;
          likewise, in the case of ngram overlap unless ALLOW_NGRAM_OVERLAP
        """
        # Get objects for top terms
        # ex: top_terms=[CorpusKeyword(term=<tfidf.dockeyword.DocKeyword object at 0x7f08b43bf550>, ngram=u'patuxent river', score=0.0015548719642054984), ... CorpusKeyword(term=<tfidf.dockeyword.DocKeyword object at 0x7f08b43cf110>, ngram=u'afognak native corporation', score=0.0009894639772216809)]
        # Get twice as many top terms to display to account for filtering
        # TODO: keep track of how often too few terms shown
        debug.trace(6, (f"get_top_terms({doc_id}, tfw:{tf_weight}, idfw:{idf_weight}, lim={limit},"
                        f"allow_sub={allow_ngram_subsumption}, allow_over={allow_ngram_overlap},"
                        f"allow_num={allow_numeric_ngrams})"))
        top_terms = self.corpus.get_keywords(document_id=doc_id,
                                             tf_weight=tf_weight,
                                             idf_weight=idf_weight,
                                             limit=(2 * limit))
        debug.trace_fmtd(7, "top_terms={tt}", tt=top_terms)
        # Skip empty tokens due to spacing and to punctuation removal (e.g, " ")
        temp_top_term_info = [(k.ngram, k.score) for k in top_terms if k.ngram.strip()]
        # Put spaces around ngrams to aid in subsumption tests
        check_ngram_overlap = (not (allow_ngram_subsumption and allow_ngram_overlap))
        if check_ngram_overlap:
            spaced_ngrams = [(" " + ngram + " ") for (ngram, _score) in temp_top_term_info]
            debug.trace_fmtd(7, "spaced_ngrams={sn}", sn=spaced_ngrams)
        top_term_info = []
        for (i, (ngram, score)) in enumerate(temp_top_term_info):
            
            if (not ngram.strip()):
                debug.trace_fmt(5, "Omitting invalid ngram '{ng}'", ng=ngram)
                continue
            ## OLD:
            ## if ((not allow_numeric_ngrams) and all([tpo.is_numeric(token) for token in ngram.split()])):
            ##     debug.trace_fmt(5, "Omitting numeric ngram '{ng}'", ng=ngram)
            ##     continue
            if ((not allow_numeric_ngrams) and any([tpo.is_numeric(token) for token in ngram.split()])):
                debug.trace_fmt(5, "Omitting ngram with numerics '{ng}'", ng=ngram)
                continue
            
            # Check for subsumption (e.g., "new york" in "new york city") and overlap (e.g. "new york" and "york city")
            ## TODO: record ngram offsets to facilitate contiguity tests
            include = True
            if check_ngram_overlap:
                for (j, other_spaced_ngram) in enumerate(spaced_ngrams):
                    is_subsumed = ((not allow_ngram_subsumption) and
                                   ((spaced_ngrams[i] in other_spaced_ngram)
                                     or (other_spaced_ngram in spaced_ngrams[i])))
                    has_overlap = ((not allow_ngram_overlap) and
                                   terms_overlap(spaced_ngrams[i], other_spaced_ngram))
                    if ((i > j) and (is_subsumed or has_overlap)):
                        include = False
                        label = ("in subsumption" if is_subsumed else "overlapping")
                        debug.trace_fmt(5, "Omitting lower-weigted ngram '{ng2}' {lbl} with '{ng1}'",
                                        ng1=other_spaced_ngram, ng2=spaced_ngrams[i], lbl=label)
                        break
            if not include:
                continue

            # OK
            top_term_info.append((ngram, score))
            if (len(top_term_info) == limit):
                break
        # Sanity check on number of terms displayed
        num_terms = len(top_term_info)
        if (num_terms < limit):
            debug.trace_fmt(3, "Warning: only {n} terms shown (of {m} max)",
                            n=num_terms, m=limit)
        debug.trace_fmtd(6, "top_term_info={tti}", tti=top_term_info)
        return top_term_info

    def old_get_ngrams(self, text):
        """Returns generator with ngrams in TEXT"""
        ## NOTE: Now returns the ngrams
        ## BAD return self.pp.yield_keywords(text)
        ngrams = []
        gen = self.pp.yield_keywords(text)
        more = True
        while (more):
            ## DEBUG: debug.trace_fmtd(6, ".")
            try:
                ## OLD: ngrams.append(gen.next().text)
                ngrams.append(next(gen).text)
            except StopIteration:
                more = False
        debug.trace_fmt(6, "ngram_tfidf_analysis.old_get_ngrams({t}) [self={s}] => {nl}", 
                        t=text, s=self, nl=ngrams)
        return ngrams

    def get_ngrams(self, text):
        """Returns ngrams in TEXT (from size MIN_NGRAM_SIZE to MAX_NGRAM_SIZE)"""
        # Based on https://stackoverflow.com/questions/13423919/computing-n-grams-using-python.
        ## OLD: vectorizer = CountVectorizer(ngram_range=(MIN_NGRAM_SIZE, MAX_NGRAM_SIZE))
        if USE_CORPUS_COUNTER:
            return self.old_get_ngrams(text)
        if self.corpus:
            ## OLD: debug.trace(2, "Warning: not using tfidf corpus object")
            debug.trace(6, "Note: not using tfidf corpus object")
        vectorizer = CountVectorizer(ngram_range=(self.min_ngram_size, self.max_ngram_size))
        analyzer = vectorizer.build_analyzer()
        ngram_list = analyzer(text)
        debug.trace_fmt(6, "ngram_tfidf_analysis.get_ngrams({t}) [self={s}] => {nl}", 
                        t=text, s=self, nl=ngram_list)
        return ngram_list

def main():
    """Entry point for script"""
    if ((len(sys.argv) == 1) or (sys.argv[1] == "--help")):
        usage = "Usage: {prog} [--help] [-]".format(prog=sys.argv[0])
        system.exit(usage)
    
    # Tabulate ngram occurrences
    ngram_analyzer = ngram_tfidf_analysis(min_ngram_size=2, max_ngram_size=3)
    all_text = system.read_entire_file(__file__)
    ## OLD: all_ngrams = ngram_analyzer.old_get_ngrams(all_text)
    all_ngrams = ngram_analyzer.get_ngrams(all_text)
    # pylint: disable=unnecessary-comprehension
    reversed_all_text = " ".join(list(reversed([token for token in all_text.split()])))
    ngram_analyzer.add_doc(all_text, doc_id="doc1")
    ngram_analyzer.add_doc(reversed_all_text, doc_id="rev-doc1")
    top_ngrams = ngram_analyzer.get_top_terms("rev-doc1", allow_ngram_subsumption=False, allow_ngram_overlap=False)

    # Check for common ngrams
    debug.assertion("simple test follows" in all_ngrams)
    debug.assertion("system getenv_boolean" in top_ngrams)
    debug.assertion("system" not in all_ngrams)
    debug.assertion("getenv_boolean" not in all_ngrams)
    
    # Check for filtering based on subsumption and overlap
    debug.assertion("warning not" in top_ngrams)
    debug.assertion("warning not intended" not in top_ngrams)

    # Check for tf/idf values
    # TODO: add assertion for specific tfidf values
    try:
        debug.assertion(ngram_analyzer.corpus.tf_idf("system getenv_boolean", document_id="doc1")
                        == ngram_analyzer.corpus.tf_idf("getenv_boolean system", document_id="rev-doc1"))
    except:
        system.print_exception_info("corpus.tf_idf")

    # Output ngram sample
    SAMPLE_SIZE = 10
    init_ngram_spec = "\n\t".join(all_ngrams[:SAMPLE_SIZE])
    print(f"first 10 ngrams in {__file__}:\n\t{init_ngram_spec}")
    init_top_ngram_spec = "\n\t".join([f"{t}: {tpo.round_num(s, 3)}"
                                       for (t, s) in top_ngrams[:SAMPLE_SIZE]])
    print(f"top ngrams in {__file__}:\n\t{init_top_ngram_spec}")

    
#-------------------------------------------------------------------------------

if __name__ == '__main__':
    system.print_stderr("Warning: not intended for command-line use; simple test follows")
    main()



#! /usr/bin/env python
#
# Copyright 2016 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
# via https://github.com/tensorflow/tensorflow/blob/master/tensorflow/python/tools/inspect_checkpoint.py

"""A simple script for inspect checkpoint files."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# TEMP:
# pylint: disable=consider-using-f-string

import argparse
import re
import sys

from absl import app
import numpy as np

# Local packages
from mezcla import debug
from mezcla.my_regex import my_re
from mezcla import system
from mezcla.text_utils import version_to_number as version_as_float


import tensorflow as tf
debug.assertion(version_as_float(tf.version.VERSION) >= version_as_float("2.0"))


from tensorflow.python.framework import errors_impl
from tensorflow.python.platform import flags
from tensorflow.python.training import py_checkpoint_reader
## import tensorflow.compat.v1 as tensorflow
## TEST
## from tensorflow.compat.v1.python.framework import errors_impl
## from tensorflow.compat.v1.python.platform import flags
## from tensorflow.compat.v1.python.training import py_checkpoint_reader


SHOW_SIZE = system.getenv_bool("SHOW_SIZE", False,
                                 "Show total count for each tensor")

FLAGS = None

def _count_total_params(reader, count_exclude_pattern=""):
    """Count total number of variables."""
    var_to_shape_map = reader.get_variable_to_shape_map()

    # Filter out tensors that we don't want to count
    if count_exclude_pattern:
        regex_pattern = re.compile(count_exclude_pattern)
        new_var_to_shape_map = {}
        exclude_num_tensors = 0
        exclude_num_params = 0
        for v in var_to_shape_map:
            if regex_pattern.search(v):
                exclude_num_tensors += 1
                exclude_num_params += np.prod(var_to_shape_map[v])
            else:
                new_var_to_shape_map[v] = var_to_shape_map[v]
            ## BAD:
            ## var_to_shape_map = new_var_to_shape_map
            ## print("# Excluding %d tensors (%d params) that match %s when counting." % (
            ##     exclude_num_tensors, exclude_num_params, count_exclude_pattern))
        var_to_shape_map = new_var_to_shape_map
        print("# Excluding %d tensors (%d params) that match %s when counting." % (
            exclude_num_tensors, exclude_num_params, count_exclude_pattern))
      
    var_sizes = [np.prod(var_to_shape_map[v]) for v in var_to_shape_map]
    return np.sum(var_sizes, dtype=int)


def add_size(listing):
    """Modify LISTING to show parameter size (e.g., arrays dimensions resolved)"""
    # example: "bert/embeddings/word_embeddings (DT_FLOAT) [30000,128]" => "...; size=3840000"
    debug.trace(6, f"add_size({listing}")
    new_lines = []
    for i, line in enumerate(listing.splitlines()):
        size = 0
        if my_re.search(r"\[(\d+),(\d+)\]", line):
            size = system.to_int(my_re.group(1)) * system.to_int(my_re.group(2))
        elif my_re.search(r"\[(\d+)\]", line):
            size = system.to_int(my_re.group(1))
        else:
            debug.trace(5, f"Ignoring listing line {i + 1}: {line}")
        line += f"; size={size}"
        new_lines.append(line)
    result = "\n".join(new_lines)
    return result


def print_tensors_in_checkpoint_file(file_name, tensor_name, all_tensors,
                                     all_tensor_names=False,
                                     count_exclude_pattern=""):
    """Prints tensors in a checkpoint file.

    If no `tensor_name` is provided, prints the tensor names and shapes
    in the checkpoint file.

    If `tensor_name` is provided, prints the content of the tensor.

    Args:
        file_name: Name of the checkpoint file.
        tensor_name: Name of the tensor in the checkpoint file to print.
        all_tensors: Boolean indicating whether to print all tensors.
        all_tensor_names: Boolean indicating whether to print all tensor names.
        count_exclude_pattern: Regex string, pattern to exclude tensors when count.
    """
    try:
        reader = py_checkpoint_reader.NewCheckpointReader(file_name)
        if all_tensors or all_tensor_names:
            var_to_shape_map = reader.get_variable_to_shape_map()
            var_to_dtype_map = reader.get_variable_to_dtype_map()
            for key, value in sorted(var_to_shape_map.items()):
                print("tensor: %s (%s) %s" % (key, var_to_dtype_map[key].name, value))
                if all_tensors:
                    try:
                        print(reader.get_tensor(key))
                    except errors_impl.InternalError:
                        print("<not convertible to a numpy dtype>")
        elif not tensor_name:
            ## OLD: print(reader.debug_string().decode("utf-8", errors="ignore"))
            listing = reader.debug_string().decode("utf-8", errors="ignore")
            if SHOW_SIZE:
                listing = add_size(listing)
            print(listing)
        else:
            if not reader.has_tensor(tensor_name):
                print("Tensor %s not found in checkpoint" % tensor_name)
                return
    
            var_to_shape_map = reader.get_variable_to_shape_map()
            var_to_dtype_map = reader.get_variable_to_dtype_map()
            print("tensor: %s (%s) %s" %
                  (tensor_name, var_to_dtype_map[tensor_name].name,
                   var_to_shape_map[tensor_name]))
            print(reader.get_tensor(tensor_name))
    
        # Count total number of parameters
        print("# Total number of params: %d" % _count_total_params(
            reader, count_exclude_pattern=count_exclude_pattern))
    except Exception as e:  # pylint: disable=broad-except
        system.print_exception_info("print_tensors_in_checkpoint_file")
        print(str(e))
        if "corrupted compressed block contents" in str(e):
            print("It's likely that your checkpoint file has been compressed "
                  "with SNAPPY.")
        if ("Data loss" in str(e) and
            any(e in file_name for e in [".index", ".meta", ".data"])):
            proposed_file = ".".join(file_name.split(".")[0:-1])
            v2_file_error_template = """
  It's likely that this is a V2 checkpoint and you need to provide the filename
  *prefix*.  Try removing the '.' and extension.  Try:
  inspect checkpoint --file_name = {}"""
            print(v2_file_error_template.format(proposed_file))
  

def parse_numpy_printoption(kv_str):
    """Sets a single numpy printoption from a string of the form 'x=y'.

    See documentation on numpy.set_printoptions() for details about what values
    x and y can take. x can be any option listed there other than 'formatter'.

    Args:
        kv_str: A string of the form 'x=y', such as 'threshold=100000'

    Raises:
        argparse.ArgumentTypeError: If the string couldn't be used to set any
            nump printoption.
    """
    k_v_str = kv_str.split("=", 1)
    if len(k_v_str) != 2 or not k_v_str[0]:
        raise argparse.ArgumentTypeError("'%s' is not in the form k=v." % kv_str)
    k, v_str = k_v_str
    printoptions = np.get_printoptions()
    if k not in printoptions:
        raise argparse.ArgumentTypeError("'%s' is not a valid printoption." % k)
    v_type = type(printoptions[k])
    if v_type is type(None):
        raise argparse.ArgumentTypeError(
            "Setting '%s' from the command line is not supported." % k)
    try:
        v = (
            v_type(v_str)
            if v_type is not bool else flags.BooleanParser().parse(v_str))
    except ValueError as e:
        raise argparse.ArgumentTypeError(e.message)
    np.set_printoptions(**{k: v})


def main(unused_argv):
    """Entry point for script"""
    if not FLAGS.file_name:
        print("Usage: inspect_checkpoint --file_name=checkpoint_file_name "
              "[--tensor_name=tensor_to_print] "
              "[--all_tensors] "
              "[--all_tensor_names] "
              "[--count_exclude_pattern regex] "
              "[--printoptions]")
        sys.exit(1)
    else:
        print_tensors_in_checkpoint_file(
            FLAGS.file_name, FLAGS.tensor_name,
            FLAGS.all_tensors, FLAGS.all_tensor_names,
            count_exclude_pattern=FLAGS.count_exclude_pattern)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.register("type", "bool", lambda v: v.lower() == "true")
    parser.add_argument(
        "--file_name",
        type=str,
        default="",
        help="Checkpoint filename. "
        "Note, if using Checkpoint V2 format, file_name is the "
        "shared prefix between all files in the checkpoint.")
    parser.add_argument(
        "--tensor_name",
        type=str,
        default="",
        help="Name of the tensor to inspect")
    parser.add_argument(
        "--count_exclude_pattern",
        type=str,
        default="",
        help="Pattern to exclude tensors, e.g., from optimizers, when counting.")
    parser.add_argument(
        "--all_tensors",
        nargs="?",
        const=True,
        type="bool",
        default=False,
        help="If True, print the names and values of all the tensors.")
    parser.add_argument(
        "--all_tensor_names",
        nargs="?",
        const=True,
        type="bool",
        default=False,
        help="If True, print the names of all the tensors.")
    parser.add_argument(
        "--printoptions",
        nargs="*",
        type=parse_numpy_printoption,
        help="Argument for numpy.set_printoptions(), in the form 'k=v'.")
    FLAGS, unparsed = parser.parse_known_args()
    app.run(main=main, argv=[sys.argv[0]] + unparsed)



#! /usr/bin/env python
#
# Functions for operating system related access, such as running command or
# getting environment values.
#
# TODO:
# - Move functions from system.py here.
#

"""OS utilities: mainly wrappers around os package"""

# Standard modules
import os

# Local modules
from mezcla import debug
from mezcla import system


def split_extension(path):
    """Returns basename and extension for PATH"""
    result = os.path.splitext(path)
    try:
        filename_proper, ext = result
        debug.assertion(system.remove_extension(path) == filename_proper)
        debug.assertion(system.get_extension(path) == ext[1:])
    except:
        pass
    debug.trace(5, f"split_extension({path}) => {result}")
    return result
